"""Added grace period column on MachineType

Revision ID: 88feb1b31a1f
Revises: 2dc275252e0e
Create Date: 2024-02-18 21:20:47.213287

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "88feb1b31a1f"
down_revision: Union[str, None] = "2dc275252e0e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("machine_types", sa.Column("grace_period_min", sa.Integer(), nullable=True))
    # ### end Alembic commands ###
    op.execute("UPDATE machine_types SET grace_period_min=2 WHERE grace_period_min IS NULL")


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("machine_types", "grace_period_min")
    # ### end Alembic commands ###



import logging
from flask_login import LoginManager, login_user, logout_user, login_required
from flask import render_template, request, redirect, url_for, flash
from .webapplication import DBSession, app
from FabOMatic.database.models import User
from FabOMatic.database.DatabaseBackend import getSetting
from flask_mail import Mail, Message
from flask_babel import gettext

login_manager = LoginManager()
login_manager.init_app(app)


app.config["MAIL_SERVER"] = getSetting("email", "server")
app.config["MAIL_PORT"] = getSetting("email", "port")
app.config["MAIL_USE_TLS"] = getSetting("email", "use_tls")
app.config["MAIL_USERNAME"] = getSetting("email", "username")
app.config["MAIL_PASSWORD"] = getSetting("email", "password")

mail = Mail(app)

SALT = b"fablab-bg"


@login_manager.user_loader
def load_user(user_id):
    with DBSession() as session:
        return session.query(User).get(int(user_id))


@app.route("/login", methods=["GET", "POST"])
def login():
    if request.method == "POST":
        with DBSession() as session:
            user = session.query(User).filter_by(email=request.form["email"]).first()
            if user and user.check_password(request.form["password"]):
                # Now check the user role
                if user.role.backend_admin or user.role.authorize_all:
                    login_user(user)
                    logging.info("User %s logged in", user.email)
                    return redirect(url_for("about"))
                else:
                    logging.warning("User %s does not have a role with backend administration permission.", user.email)
                    flash(gettext("Your user does not have a role with backend administration permission."), "danger")
                    return redirect(url_for("login"))
            else:
                logging.warning("Failed login attempt for user %s", request.form["email"])
                flash(gettext("Wrong username or password."), "danger")
                return redirect(url_for("login"))
    return render_template("login.html")


@app.route("/logout")
@login_required
def logout():
    logout_user()
    flash(gettext("You have been logged out."), "success")
    return redirect(url_for("login"))


def send_reset_email(user: User) -> bool:
    token = user.get_reset_token(app.config["SECRET_KEY"], SALT)
    msg = Message("Password Reset Request", sender="admin@fablab.org", recipients=[user.email])
    msg.body = f"""To reset your password, visit the following link:
                {url_for('reset_token', token=token, _external=True)}

                If you did not make this request then simply ignore this email and no changes will be made.
                """
    mail.send(msg)
    return True


@app.route("/forgot_password", methods=["GET", "POST"])
def forgot_password():
    if request.method == "POST":
        with DBSession() as session:
            user = session.query(User).filter_by(email=request.form["email"]).first()
            if user:
                send_reset_email(user)
                flash(gettext("Email sent with instructions to reset your password."), "info")
                return redirect(url_for("login"))
            else:
                flash(gettext("No user found with this email."), "danger")
                return redirect(url_for("login"))
    return render_template("forgot_password.html")


@app.route("/reset_token/<token>", methods=["GET", "POST"])
def reset_token(token):
    user_id = User.verify_reset_token(token, app.config["SECRET_KEY"], SALT)
    if not user_id:
        flash(gettext("That is an invalid or expired token"), "warning")
        return redirect(url_for("forgot_password"))
    if request.method == "POST":
        with DBSession() as session:
            user = session.query(User).get(user_id)
            user.set_password(request.form["password"])
            session.commit()
            flash(gettext("Your password has been updated! You are now able to log in"), "success")
            return redirect(url_for("login"))
    return render_template("reset_token.html", title="Reset Password")



"""Machine types: added possibility to disable authorization

Revision ID: dfd2b0db0016
Revises: aa9ed3e094d5
Create Date: 2024-05-05 20:27:50.705509

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'dfd2b0db0016'
down_revision: Union[str, None] = 'aa9ed3e094d5'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('machine_types', sa.Column('access_management', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('machine_types', 'access_management')
    # ### end Alembic commands ###



""" This module contains the MachineLogic class. """

import logging

from time import time

from FabOMatic.mqtt.mqtt_types import AliveQuery, MachineResponse, SimpleResponse, UserResponse
from FabOMatic.database.DatabaseBackend import DatabaseBackend
from FabOMatic.database.constants import DEFAULT_GRACE_PERIOD_MINUTES, DEFAULT_TIMEOUT_MINUTES, USER_LEVEL


class MachineLogic:
    """
    The MachineLogic class represents the logic for interacting with a machine in the system.
    It provides methods for checking the machine status, handling machine alive messages,
    authorizing users, starting and ending machine use, and registering maintenance interventions.
    """

    database: DatabaseBackend = None

    def __init__(self, machine_id: int):
        """
        Initializes a new instance of the MachineLogic class.

        Args:
            machine_id (int): The ID of the machine.

        Raises:
            Exception: If the database is not initialized.
            Exception: If the machine ID is invalid.
        """
        self._machine_id = machine_id
        self._last_alive = 0

        if MachineLogic.database is None:
            raise ValueError("Database not initialized")

        with self.database.getSession() as session:
            machine_repo = MachineLogic.database.getMachineRepository(session)
            if machine_repo.get_by_id(machine_id) is None:
                raise ValueError("Invalid machine id")

        logging.info(f"Machine logic instance for ID:{machine_id} initialized")

    def updateMachineLastSeen(self):
        """
        Updates the last seen timestamp of the machine.
        """
        self._last_alive = time()
        try:
            with MachineLogic.database.getSession() as session:
                machine_repo = MachineLogic.database.getMachineRepository(session)
                machine = machine_repo.get_by_id(self._machine_id)
                machine.last_seen = time()
                machine_repo.update(machine)
        except Exception as e:
            logging.error("updateMachineLastSeen exception %s", str(e), exc_info=True)

    def machineStatus(self):
        """
        Gets the status of the machine.

        Returns:
            MachineResponse: The machine response object containing the status information.
        """
        try:
            with MachineLogic.database.getSession() as session:
                machine_repo = MachineLogic.database.getMachineRepository(session)
                machine = machine_repo.get_by_id(self._machine_id)
                if machine is None:
                    return MachineResponse(
                        True, False, False, False, "?", 0, DEFAULT_TIMEOUT_MINUTES, DEFAULT_GRACE_PERIOD_MINUTES, ""
                    )
                self.updateMachineLastSeen()
                maintenance = machine_repo.getMachineMaintenanceNeeded(machine.machine_id)
                return MachineResponse(
                    True,
                    True,
                    maintenance[0],
                    not machine.blocked,
                    machine.machine_name,
                    machine.machine_type_id,
                    machine.machine_type.type_timeout_min,
                    machine.machine_type.grace_period_min,
                    maintenance[1],
                )
        except Exception as e:
            logging.error("machineStatus exception %s", str(e), exc_info=True)
            return MachineResponse(
                False, False, False, False, "?", 0, DEFAULT_TIMEOUT_MINUTES, DEFAULT_GRACE_PERIOD_MINUTES, ""
            )

    def machineAlive(self, alive: AliveQuery):
        """
        Called when a machine sends an alive message.
        """
        with MachineLogic.database.getSession() as session:
            board_repo = MachineLogic.database.getBoardsRepository(session)
            machine = MachineLogic.database.getMachineRepository(session).get_by_id(self._machine_id)
            board_repo.registerBoard(alive.ip, alive.version, alive.serial, alive.heap, machine)
            self.updateMachineLastSeen()

    def isAuthorized(self, card_uuid: str) -> UserResponse:
        """
        Checks if a user is authorized to use the machine.

        Args:
            card_uuid (str): The UUID of the user's card.

        Returns:
            UserResponse: The user response object containing the authorization information.
        """
        try:
            self.updateMachineLastSeen()
            with MachineLogic.database.getSession() as session:
                machine_repo = MachineLogic.database.getMachineRepository(session)
                user_repo = MachineLogic.database.getUserRepository(session)
                user = user_repo.getUserByCardUUID(card_uuid)
                machine = machine_repo.get_by_id(self._machine_id)
                if user is None and machine is not None:
                    unknown_repo = MachineLogic.database.getUnknownCardsRepository(session)
                    unknown_repo.registerUnknownCard(card_uuid, machine)

                if machine is None or user is None:
                    return UserResponse(True, False, "Unknown", USER_LEVEL.INVALID, False)
                if user_repo.IsUserAuthorizedForMachine(machine, user):
                    return UserResponse(True, True, user.name, user.user_level(), False)
                else:
                    return UserResponse(True, False, "User not authorized", USER_LEVEL.INVALID, True)

        except Exception as e:
            logging.error("isAuthorized exception %s", str(e), exc_info=True)
            return UserResponse(False, False, "", USER_LEVEL.INVALID, True)

    def startUse(self, card_uuid: str, replay: bool) -> SimpleResponse:
        """
        Starts the use of the machine by a user.

        Args:
            card_uuid (str): The UUID of the user's card.

        Returns:
            SimpleResponse: The simple response object indicating the success or failure of the operation.
        """
        try:
            self.updateMachineLastSeen()
            with MachineLogic.database.getSession() as session:
                user_repo = MachineLogic.database.getUserRepository(session)
                user = user_repo.getUserByCardUUID(card_uuid)
                if user is None:
                    return SimpleResponse(False, "Invalid card")

                use_repo = MachineLogic.database.getUseRepository(session)
                result = use_repo.startUse(self._machine_id, user, time(), replay)

                return SimpleResponse(result, "")
        except Exception as e:
            logging.error("startUse exception %s", str(e), exc_info=True)
            return SimpleResponse(False, "BACKEND EXCEPTION")

    def inUse(self, card_uuid: str, duration_s: int) -> SimpleResponse:
        """
        Update current usage of the machine by a user.

        Args:
            card_uuid (str): The UUID of the user's card.
            duration_s (int): The duration of the machine use in seconds.

        Returns:
            SimpleResponse: The simple response object indicating the success or failure of the operation.
        """
        try:
            self.updateMachineLastSeen()
            with MachineLogic.database.getSession() as session:
                user_repo = MachineLogic.database.getUserRepository(session)
                user = user_repo.getUserByCardUUID(card_uuid)
                if user is None:
                    return SimpleResponse(False, "Invalid card")

                use_repo = MachineLogic.database.getUseRepository(session)
                result = use_repo.inUse(self._machine_id, user, duration_s)

                return SimpleResponse(result, "inUse")
        except Exception as e:
            logging.error("inuse exception %s", str(e), exc_info=True)
            return SimpleResponse(False, "BACKEND EXCEPTION")

    def endUse(self, card_uuid: str, duration_s: int, replay: bool) -> SimpleResponse:
        """
        Ends the use of the machine by a user.

        Args:
            card_uuid (str): The UUID of the user's card.
            duration_s (int): The duration of the machine use in seconds.

        Returns:
            SimpleResponse: The simple response object indicating the success or failure of the operation.
        """
        try:
            self.updateMachineLastSeen()
            with MachineLogic.database.getSession() as session:
                user_repo = MachineLogic.database.getUserRepository(session)
                user = user_repo.getUserByCardUUID(card_uuid)
                if user is None:
                    return SimpleResponse(False, "Invalid card")

                use_repo = MachineLogic.database.getUseRepository(session)
                duration_s = use_repo.endUse(self._machine_id, user, duration_s, replay)

                return SimpleResponse(True, f"Duration {duration_s} seconds")
        except Exception as e:
            logging.error("enduse exception %s", str(e), exc_info=True)
            return SimpleResponse(False, "BACKEND EXCEPTION")

    def registerMaintenance(self, card_uuid: str, replay: bool) -> SimpleResponse:
        """
        Registers a maintenance intervention for the machine.

        Args:
            card_uuid (str): The UUID of the user's card.
            replay (bool): If this message has been buffered and sent later on

        Returns:
            SimpleResponse: The simple response object indicating the success or failure of the operation.
        """
        try:
            self.updateMachineLastSeen()
            with MachineLogic.database.getSession() as session:
                user_repo = MachineLogic.database.getUserRepository(session)
                user = user_repo.getUserByCardUUID(card_uuid)
                if user is None:
                    return SimpleResponse(False, "Wrong user card")
                if not user.role.maintenance or user.disabled:
                    return SimpleResponse(False, "Not authorized")

                intervention_repo = MachineLogic.database.getInterventionRepository(session)
                intervention_repo.registerInterventionsDone(self._machine_id, user.user_id)

                return SimpleResponse(True, "")
        except Exception as e:
            logging.error("registerMaintenance exception %s", str(e), exc_info=True)
            return SimpleResponse(False, "BACKEND EXCEPTION")

    def getMachineId(self) -> int:
        return self._machine_id



""" Database repositories for the FabOMatic application."""

import logging

from time import time
from typing import List, Optional, Tuple

from sqlalchemy.orm import Session

from .models import (
    Board,
    Use,
    Machine,
    MachineType,
    Intervention,
    User,
    Maintenance,
    Authorization,
    Role,
    Base,
    UnknownCard,
)
from .constants import DEFAULT_TIMEOUT_MINUTES
from sqlalchemy.orm import object_session


class BaseRepository:
    """Base class for all repositories."""

    def __init__(self, db_session: Session):
        self.db_session = db_session

    def create(self, model: Base):
        """Create a new model instance in the database."""
        self.db_session.add(model)
        self.db_session.commit()
        self.db_session.refresh(model)

    def update(self, model: Base):
        """Update an existing model instance in the database."""
        if object_session(model) is None:
            raise ValueError("Object is not bound to a session")
        object_session(model).commit()
        self.db_session.refresh(model)

    def delete(self, model: Base):
        """Delete a model instance from the database."""
        self.db_session.delete(model)
        self.db_session.commit()

    def rollback(self):
        """Rollback the current database session."""
        self.db_session.rollback()

    def get_all_model(self, model_class: type) -> List[Base]:
        """Retrieve all model instances of a given class from the database."""
        return self.db_session.query(model_class).all()

    def filter_by_model(self, model_class: type, **kwargs) -> List[Base]:
        """Filter model instances of a given class by specified criteria."""
        return self.db_session.query(model_class).filter_by(**kwargs).all()

    def get_model_by_id(self, model_class: type, id_filter: int) -> Base:
        """Retrieve a model instance of a given class by its ID."""
        primary_key_name = model_class.__table__.primary_key.columns.keys()[0]
        return self.db_session.query(model_class).filter_by(**{primary_key_name: id_filter}).first()

    def bulk_create(self, entities):
        self.db_session.bulk_save_objects(entities)


class RoleRepository(BaseRepository):
    """
    Repository class for managing Role objects in the database.
    """

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_by_role_name(self, role_name: str) -> Optional[Role]:
        """
        Retrieve a role by its name.

        Args:
            role_name (str): The name of the role.

        Returns:
            Optional[Role]: The role object if found, None otherwise.
        """
        return self.db_session.query(Role).filter_by(role_name=role_name).first()

    def get_all(self) -> List[Role]:
        """
        Retrieve all roles.

        Returns:
            List[Role]: A list of all roles.
        """
        return self.db_session.query(Role).order_by(Role.role_id).all()

    def get_by_id(self, id: int) -> Base:
        """
        Retrieve a role by its ID.

        Args:
            id (int): The ID of the role.

        Returns:
            Base: The role object if found, None otherwise.
        """
        return self.db_session.query(Role).filter_by(role_id=id).first()


class MachineTypeRepository(BaseRepository):
    """
    Repository class for managing MachineType entities.
    """

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_all(self) -> List[MachineType]:
        """
        Retrieves all MachineType entities from the database.

        Returns:
            A list of MachineType entities.
        """
        return self.db_session.query(MachineType).order_by(MachineType.type_id).all()

    def get_by_id(self, id: int) -> Base:
        """
        Retrieves a MachineType entity by its ID.

        Args:
            id: The ID of the MachineType entity to retrieve.

        Returns:
            The MachineType entity with the specified ID, or None if not found.
        """
        return self.db_session.query(MachineType).filter_by(type_id=id).first()


class UserRepository(BaseRepository):
    """Repository class for managing User objects in the database."""

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def getUserByCardUUID(self, card_uuid: str) -> Optional[User]:
        """Get a user by their card UUID.

        Args:
            card_uuid (str): The card UUID of the user.

        Returns:
            Optional[User]: The user object if found, None otherwise.
        """
        return self.db_session.query(User).filter_by(card_UUID=card_uuid).first()

    def getUserUses(self, user: User) -> List[Use]:
        """Get all the uses of a user.

        Args:
            user (User): The user object.

        Returns:
            List[Use]: A list of use objects associated with the user.
        """
        return self.db_session.query(Use).filter(Use.user_id == user.user_id).all()

    def getUserTotalTime(self, user_id: int) -> float:
        """Return total time the User has used the machines.

        Args:
            user_id (int): id of the User

        Returns:
            float: User time in seconds
        """
        uses = self.db_session.query(Use).filter(Use.user_id == user_id, Use.end_timestamp.is_not(None)).all()
        return sum([use.end_timestamp - use.start_timestamp for use in uses], 0)

    def IsUserAuthorizedForMachine(self, machine: Machine, user: User) -> bool:
        """Return True if the User is authorized to use the Machine, False otherwise.

        Args:
            machine (Machine): The machine object.
            user (User): The user object.

        Returns:
            bool: True if the user is authorized, False otherwise.
        """
        if user.disabled:
            return False
        if user.deleted:
            return False

        if user.role.authorize_all:
            return True

        if machine.blocked:
            return False

        if machine.machine_type.access_management == MachineType.MANAGEMENT_WITHOUT_AUTHORIZATION:
            return True

        authorizations = self.db_session.query(Authorization).filter_by(user_id=user.user_id).all()
        machine_ids = [a.machine_id for a in authorizations]

        return machine.machine_id in machine_ids

    def get_all(self) -> List[User]:
        """Get all users.

        Returns:
            List[User]: A list of all user objects.
        """
        return self.db_session.query(User).order_by(User.user_id).all()

    def get_by_id(self, id: int) -> Optional[User]:
        """Get a user by their ID.

        Args:
            id (int): The ID of the user.

        Returns:
            Optional[User]: The user object if found, None otherwise.
        """
        return self.db_session.query(User).filter_by(user_id=id).first()

    def get_anonymous(self) -> Optional[User]:
        """Get the anonymous user by their ID.

        Returns:
            Optional[User]: The user object if found, None otherwise.
        """
        role = self.db_session.query(Role).filter_by(role_name="Anonymous").first()
        if role is None:
            return None

        return self.db_session.query(User).filter_by(role_id=role.role_id).first()


class AuthorizationRepository(BaseRepository):
    """
    Repository class for managing authorizations in the database.
    """

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_all(self) -> List[Authorization]:
        """
        Retrieves all authorizations from the database.

        Returns:
            A list of Authorization objects.
        """
        return self.db_session.query(Authorization).order_by(Authorization.authorization_id).all()


class MaintenanceRepository(BaseRepository):
    """
    Repository class for handling maintenance operations.
    """

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_all(self) -> List[Maintenance]:
        """
        Retrieve all maintenance records from the database.

        Returns:
            A list of Maintenance objects.
        """
        return self.db_session.query(Maintenance).order_by(Maintenance.maintenance_id).all()


class InterventionRepository(BaseRepository):
    def __init__(self, db_session: Session):
        """
        Initializes an instance of InterventionRepository.

        Args:
            db_session (Session): The database session object.
        """
        super().__init__(db_session)

    def get_all(self) -> List[Intervention]:
        """
        Retrieves all interventions from the database.

        Returns:
            List[Intervention]: A list of Intervention objects.
        """
        return self.db_session.query(Intervention).order_by(Intervention.intervention_id).all()

    def registerInterventionsDone(self, machine_id: int, user_id: int):
        """
        Register interventions done by the user on the machine.

        Args:
            machine_id (int): The ID of the machine.
            user_id (int): The ID of the user.

        Raises:
            ValueError: If the machine_id or user_id is invalid.
            ValueError: If the user is disabled.
        """
        machine = self.db_session.query(Machine).filter_by(machine_id=machine_id).first()
        user = self.db_session.query(User).filter_by(user_id=user_id).first()

        if machine is None or user is None:
            raise ValueError("Wrong machine_id or user_id")

        if user.disabled:
            raise ValueError("Invalid user")

        machine_repo = MachineRepository(self.db_session)
        timestamp = time()
        for maintenance in machine.maintenances:
            if (
                machine_repo.getRelativeUseTimeByMaintenance(machine.machine_id, maintenance)
                > maintenance.hours_between * 3600
            ):
                intervention = Intervention(
                    machine_id=machine.machine_id,
                    maintenance_id=maintenance.maintenance_id,
                    timestamp=timestamp,
                    user_id=user.user_id,
                )
                self.db_session.add(intervention)
        self.db_session.commit()


class MachineRepository(BaseRepository):
    """
    Repository class for interacting with the Machine table in the database.
    """

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def isMachineCurrentlyUsed(self, machine_id: int) -> bool:
        """Return True if the machine is currently being used, False otherwise.

        Args:
            machine_id (int): id of the Machine

        Returns:
            bool
        """
        return (
            self.db_session.query(Use).filter(Use.machine_id == machine_id, Use.end_timestamp.is_(None)).first()
            is not None
        )

    def getTimeout(self, machine_id: int) -> int:
        """
        Retrieves the timeout value for a given machine.

        Args:
            machine_id (int): The ID of the machine.

        Returns:
            int: The timeout value in minutes.
        """
        machine = self.get_by_id(machine_id)
        if machine is None:
            return DEFAULT_TIMEOUT_MINUTES
        return machine.machine_type.type_timeout_min

    def getCurrentlyUsedMachines(self) -> List[Machine]:
        """Get a list of the Machine that are being used in this moment.

        Returns:
            List[Machine]
        """
        return (
            self.db_session.query(Machine)
            .join(Use, Machine.machine_id == Use.machine_id)
            .filter(Use.end_timestamp.is_(None))
            .all()
        )

    def get_by_id(self, machine_id: int = None) -> Optional[Machine]:
        """
        Retrieve a Machine by its ID.

        Args:
            machine_id (int): The ID of the machine.

        Returns:
            Optional[Machine]: The Machine object if found, None otherwise.
        """
        return self.db_session.query(Machine).filter_by(machine_id=machine_id).first()

    def get_all(self) -> List[Machine]:
        """
        Retrieve all machines.

        Returns:
            List[Machine]: A list of all machines.
        """
        return self.db_session.query(Machine).order_by(Machine.machine_id).all()

    def getRelativeUseTime(self, machine_id: int) -> int:
        """Return total time the Machine has been used since last intervention.

        Args:
            machine_id (int): id of the Machine

        Returns:
            int: Machine time in seconds
        """
        record = (
            self.db_session.query(Intervention)
            .filter(Intervention.machine_id == machine_id)
            .order_by(Intervention.timestamp.desc())
            .first()
        )

        if record is None:
            last_intervention = 0
        else:
            last_intervention = record.timestamp

        uses = (
            self.db_session.query(Use)
            .filter(
                Use.end_timestamp.is_not(None), Use.machine_id == machine_id, Use.start_timestamp > last_intervention
            )
            .all()
        )

        return sum([use.end_timestamp - use.start_timestamp for use in uses], 0)

    def getRelativeUseTimeByMaintenance(self, machine_id: int, maintenance: Maintenance) -> int:
        """Return total time the Machine has been used since last intervention.

        Args:
            machine_id (int): id of the Machine
            maintenance_id (int): id of the Maintenance

        Returns:
            int: Machine time in seconds
        """
        record = (
            self.db_session.query(Intervention)
            .filter(Intervention.machine_id == machine_id, Intervention.maintenance_id == maintenance.maintenance_id)
            .order_by(Intervention.timestamp.desc())
            .first()
        )

        if record is None:
            last_intervention = 0
        else:
            last_intervention = record.timestamp

        uses = (
            self.db_session.query(Use)
            .filter(
                Use.end_timestamp.is_not(None), Use.machine_id == machine_id, Use.start_timestamp > last_intervention
            )
            .all()
        )

        relative_time = sum([use.end_timestamp - use.start_timestamp for use in uses], 0)
        logging.debug(
            f"Machine {machine_id} Maintenance [{maintenance.description} - {maintenance.lcd_message}], last_intervention {last_intervention} relative time = {relative_time} seconds ({len(uses)} uses)"
        )
        return relative_time

    def getTotalUseTime(self, machine_id: int) -> int:
        """Return total time the Machine has been used.

        Args:
            machine_id (int): id of the Machine

        Returns:
            int: Machine time in seconds
        """
        uses = self.db_session.query(Use).filter(Use.end_timestamp.is_not(None), Use.machine_id == machine_id).all()

        total_time = sum([use.end_timestamp - use.start_timestamp for use in uses], 0)
        logging.debug(f"Machine {machine_id} total time = {total_time} seconds ({len(uses)} uses)")
        return total_time

    def getMachineMaintenanceNeeded(self, machine_id: int) -> Tuple[bool, str]:
        """Return True if the Machine needs any maintenance, False otherwise.

        Args:
            machine_id (int): id of the Machine

        Returns:
            bool
        """
        machine = self.db_session.query(Machine).filter(Machine.machine_id == machine_id).first()
        for maintenance in machine.maintenances:
            if self.getRelativeUseTimeByMaintenance(machine_id, maintenance) > maintenance.hours_between * 3600.0:
                logging.debug(f"Machine {machine_id} needs maintenance [{maintenance.description}]")
                return (True, maintenance.lcd_message)

        return (False, "")


class UseRepository(BaseRepository):
    MAX_DURATION = 3600 * 24

    """Repository class for managing Use objects in the database."""

    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def startUse(self, machine_id: int, user: User, timestamp: float = None, is_replay: bool = False) -> bool:
        """Start a new Use of a Machine by a User.

        Args:
            machine_id (int): id of the Machine
            user (User): User that is using the Machine
            timestamp (float, optional): timestamp of the start of the Use. Current time if None. Defaults to None.

        Returns:
            bool: True if the use was started successfully, False otherwise.
        """

        if timestamp is None:
            timestamp = time()

        machine_repo = MachineRepository(self.db_session)
        machine = machine_repo.get_by_id(machine_id)
        if machine is None or user is None:
            return False

        # Close eventual previous uses which were not closed with duration == (last_seen - start_timestamp)
        for rec in self.db_session.query(Use).filter(
            Use.machine_id == machine.machine_id, Use.end_timestamp.is_(None)
        ):
            logging.warning("Missed stopUse event in previous record, forcing stopUse")
            self.endUse(rec.machine_id, rec.user, int(rec.last_seen - rec.start_timestamp), is_replay)

        # Register start of new use
        self.db_session.add(
            Use(
                user_id=user.user_id,
                machine_id=machine.machine_id,
                start_timestamp=timestamp,
                last_seen=timestamp,
                replay=is_replay,
            )
        )
        self.db_session.commit()

        return True

    def inUse(self, machine_id: int, user: User, duration_s: int) -> bool:
        """Register current usage of the machine by a user.

        Args:
            machine_id (int): id of the Machine
            user (User): User that is using the Machine.
            duration_s (int): duration of the Use.

        Returns:
            bool: if the inUse was registered successfully.
        """
        machine_repo = MachineRepository(self.db_session)
        machine = machine_repo.get_by_id(machine_id)
        if machine is None or user is None or duration_s < 0 or duration_s > UseRepository.MAX_DURATION:
            return False

        record = self.db_session.query(Use).filter(Use.machine_id == machine_id, Use.end_timestamp.is_(None)).first()
        end = time()

        if record is None:
            # InUse received but no startUse was received before
            record = Use(
                machine_id=machine_id,
                user_id=user.user_id,
                start_timestamp=(end - duration_s),
                last_seen=end,
                end_timestamp=None,
            )
            logging.warning("Missed startUse event, using inUse data.")
            self.create(record)
        else:
            # Update existing record
            record.last_seen = end

        self.db_session.commit()

        return True

    def endUse(self, machine_id: int, user: User, duration_s: int, is_replay: bool) -> int:
        """End a use that was previously started.

        Args:
            machine_id (int): id of the Machine
            user (User): User that is using the Machine.
            duration_s (int): duration of the Use.

        Returns:
            int: duration of the Use in seconds
        """
        machine_repo = MachineRepository(self.db_session)
        machine = machine_repo.get_by_id(machine_id)
        if machine is None or user is None:
            return 0

        if duration_s < 0 or duration_s > UseRepository.MAX_DURATION:
            duration_s = 1

        record = self.db_session.query(Use).filter(Use.machine_id == machine_id, Use.end_timestamp.is_(None)).first()
        end = time()
        if record is None:
            # Create missing record on the fly since we have all required information
            record = Use(
                machine_id=machine_id,
                user_id=user.user_id,
                start_timestamp=(end - duration_s),
                last_seen=end,
                end_timestamp=end,
                replay=is_replay,
            )

            # Check that there is no duplicate (could happen if a client sends several identical stopUse requests)
            existing_record = (
                self.db_session.query(Use)
                .filter(
                    Use.machine_id == record.machine_id,
                    Use.user_id == record.user_id,
                    Use.start_timestamp == record.start_timestamp,
                    Use.end_timestamp == record.end_timestamp,
                )
                .first()
            )
            if existing_record is None:
                logging.warning("Missing startUse detected, creating new record on the fly.")
                self.create(record)
                machine.machine_hours += (record.end_timestamp - record.start_timestamp) / 3600.0
            else:
                logging.warning("Duplicate stopUse detected, ignoring client request")
        else:
            # Update existing record
            record.end_timestamp = record.start_timestamp + duration_s
            record.last_seen = end
            self.db_session.commit()

            # Close eventual previous uses which were not closed
            for rec in self.db_session.query(Use).filter(Use.machine_id == machine_id, Use.end_timestamp.is_(None)):
                duration_s += int(rec.last_seen - rec.start_timestamp)
                rec.end_timestamp = rec.last_seen

        self.db_session.commit()

        machine.machine_hours += duration_s / 3600.0
        machine_repo.update(machine)

        return duration_s

    def get_all(self) -> List[Use]:
        """
        Retrieve all Use objects from the database.

        Returns:
            List[Use]: A list of all Use objects.
        """
        return self.db_session.query(Use).order_by(Use.use_id).all()


class UnknownCardsRepository(BaseRepository):
    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_all(self) -> List[UnknownCard]:
        return self.db_session.query(UnknownCard).order_by(UnknownCard.id).all()

    def registerUnknownCard(self, uuid: str, machine: Machine) -> int:
        """Register an unknown card that was used on a machine.

        Args:
            card_uuid (str): UUID of the unknown card
            machine (Machine): Machine where the card was used

        Returns:
            int: id of the new UnknownCard record
        """
        record = UnknownCard(card_UUID=uuid, machine_id=machine.machine_id, timestamp=time())
        self.create(record)
        self.db_session.commit()
        return record.id


class BoardsRepository(BaseRepository):
    def __init__(self, db_session: Session):
        super().__init__(db_session)

    def get_all(self) -> List[Board]:
        return self.db_session.query(Board).order_by(Board.board_id).all()

    def registerBoard(self, ip: str, version: str, serial: str, heap: int, machine: Machine) -> int:
        """Register a board that was used on a machine.

        Args:
            ip (str): IP address of the board
            version (str): Firmware version of the board
            serial (str): Serial number of the board
            heap (int): Free heap size of the board
            machine (Machine): Machine where the board was used

        Returns:
            int: id of the new Board record
        """
        if machine is None:
            raise ValueError("Invalid machine")

        # Checks if a board already exists for this machine
        record = self.db_session.query(Board).filter(Board.machine_id == machine.machine_id).first()
        if record is not None:
            record.ip_address = ip
            record.fw_version = version
            record.serial = serial
            record.heap = heap
            record.last_seen = time()
            self.update(record)
            self.db_session.commit()
            logging.debug(
                f"Updated board #{record.board_id} for machine {machine.machine_id} (IP: {ip}, FW: {version}, Serial: {serial})"
            )
            return record.board_id

        # Create a new record
        record = Board(
            ip_address=ip,
            fw_version=version,
            machine_id=machine.machine_id,
            serial=serial,
            heap=heap,
            last_seen=time(),
        )
        self.create(record)
        self.db_session.commit()
        logging.info(
            f"Registered new board #{record.board_id} for machine {machine.machine_id} (IP: {ip}, FW: {version})"
        )
        return record.board_id



"""This is the class handling the Database. More to come."""

import os
from os import path
from os.path import dirname, abspath
import logging
import toml
from sqlalchemy import create_engine
from sqlalchemy.orm.exc import NoResultFound

from FabOMatic.database.models import MachineType, Role, User, Machine, Maintenance

from .repositories import (
    BoardsRepository,
    UseRepository,
    UserRepository,
    RoleRepository,
    MachineRepository,
    MachineTypeRepository,
    AuthorizationRepository,
    MaintenanceRepository,
    InterventionRepository,
    UnknownCardsRepository,
    Session,
    Base,
)

MODULE_DIR = dirname(dirname(abspath(__file__)))
CONFIG_FILE = path.join(MODULE_DIR, "conf", "settings.toml")
MIGRATIONS_DIR = path.join(MODULE_DIR, "alembic")


def getSetting(section: str, setting: str, settings_path: str = CONFIG_FILE) -> str:
    """Return setting from settings.toml.

    Args:
        setting (str): Setting to return
        section (str): Section of setting
        settings_path (str, optional): Path to settings.toml. Defaults to CONFIG_FILE.
    """
    settings = toml.load(settings_path)
    return settings[section][setting]


def getDatabaseUrl(settings_path: str = CONFIG_FILE) -> str:
    # Get the database URL
    db_url = getSetting("database", "url", settings_path)

    # Check if it's a SQLite URL
    if db_url.startswith("sqlite:///"):
        # Remove the prefix to get the file path
        file_path = db_url[len("sqlite:///") :]

        # Get the absolute path
        absolute_path = os.path.abspath(file_path)

        # Add the prefix back to get the absolute URL
        absolute_url = "sqlite:///" + absolute_path
    else:
        # If it's not a SQLite URL, just use it as is
        absolute_url = db_url
    return absolute_url


class DatabaseBackend:
    """Class handling the connection from and to the database."""

    def __init__(self, settings_path=CONFIG_FILE) -> None:
        """Create instance of Database.

        Args:
            settings_path (str, optional): TOML file settings. Defaults to CONFIG_FILE
        """
        self._settings_path = settings_path
        self._settings = None

        self._loadSettings()
        self._connect()

    def _loadSettings(self) -> None:
        """Load settings from TOML file."""
        self._settings = toml.load(self._settings_path)
        self._url = getDatabaseUrl(self._settings_path)
        self._name = self._settings["database"]["name"]

    def _connect(self) -> None:
        """Connect to the database."""
        from sqlalchemy.orm import sessionmaker

        self._engine = create_engine(self._url, echo=False)
        self._session = sessionmaker(bind=self._engine)
        logging.info("Connected to database %s", self._url)

    def getOne(self, Model, **kwargs):
        """Return one instance of Model matching kwargs.

        Args:
            Model: The model class to query.
            **kwargs: Keyword arguments to filter the query.

        Returns:
            The first instance of Model matching the kwargs, or None if not found.
        """
        with self._session() as session:
            try:
                result = session.query(Model).filter_by(**kwargs).one()
                return result
            except NoResultFound:
                return None

    def query(self, Model, **kwargs):
        """Query for instances of Model matching kwargs.

        Args:
            Model: The model class to query.
            **kwargs: Keyword arguments to filter the query.

        Returns:
            A list of instances of Model matching the kwargs.
        """
        with self._session() as session:
            return session.query(Model).filter_by(**kwargs).all()

    def getSession(self) -> Session:
        """Get a session object for database operations.

        Returns:
            A session object.
        """
        return self._session()

    def getUserRepository(self, session: Session) -> UserRepository:
        """Get a UserRepository object for user-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A UserRepository object.
        """
        return UserRepository(session)

    def getRoleRepository(self, session: Session) -> RoleRepository:
        """Get a RoleRepository object for role-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A RoleRepository object.
        """
        return RoleRepository(session)

    def getMachineRepository(self, session: Session) -> MachineRepository:
        """Get a MachineRepository object for machine-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A MachineRepository object.
        """
        return MachineRepository(session)

    def getMachineTypeRepository(self, session: Session) -> MachineTypeRepository:
        """Get a MachineTypeRepository object for machine type-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A MachineTypeRepository object.
        """
        return MachineTypeRepository(session)

    def getUseRepository(self, session: Session) -> UseRepository:
        """Get a UseRepository object for use-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A UseRepository object.
        """
        return UseRepository(session)

    def getAuthorizationRepository(self, session: Session) -> AuthorizationRepository:
        """Get an AuthorizationRepository object for authorization-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            An AuthorizationRepository object.
        """
        return AuthorizationRepository(session)

    def getMaintenanceRepository(self, session: Session) -> MaintenanceRepository:
        """Get a MaintenanceRepository object for maintenance-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            A MaintenanceRepository object.
        """
        return MaintenanceRepository(session)

    def getInterventionRepository(self, session: Session) -> InterventionRepository:
        """Get an InterventionRepository object for intervention-related database operations.

        Args:
            session: The session object to use for database operations.

        Returns:
            An InterventionRepository object.
        """
        return InterventionRepository(session)

    def getUnknownCardsRepository(self, session: Session) -> UnknownCardsRepository:
        return UnknownCardsRepository(session)

    def getBoardsRepository(self, session: Session) -> BoardsRepository:
        return BoardsRepository(session)

    def deleteExistingDatabase(self) -> bool:
        """Delete the existing database."""
        if self._url.startswith("sqlite:///"):
            file_path = self._url[len("sqlite:///") :]
            if os.path.exists(file_path):
                try:
                    os.remove(file_path)
                except Exception as e:
                    logging.error("Error deleting existing database %s: %s", file_path, e)
                    return False
                logging.warning("Deleted existing database %s", file_path)
                return True
            else:
                logging.warning("No existing database found at %s", file_path)
                return False
        logging.warning("Not deleting non-SQLite database %s", self._url)
        return False

    def createAndUpdateDatabase(self) -> None:
        """Create the database."""
        # Runs any migrations
        from alembic.config import Config
        from alembic import command

        alembic_cfg = Config("alembic.ini")
        alembic_cfg.set_main_option("sqlalchemy.url", self._url)
        alembic_cfg.set_main_option("script_location", MIGRATIONS_DIR)
        alembic_cfg.attributes["configure_logger"] = False
        command.upgrade(alembic_cfg, "head")

        with self._session() as session:
            if len(self.getUserRepository(session).get_all()) == 0:
                self.seedDatabase()

        if self.getUserRepository(session).get_anonymous() is None:
            logging.warning("Adding anonymous role and user to existing database")
            role = self.seedAnonymousRole(session)
            self.seedAnonymousUser(session, role)

    def seedDatabase(self) -> None:
        """Seed the database with initial data."""
        logging.warning("Seeding empty database %s", self._url)
        with self._session() as session:
            mt1 = MachineType(type_name="Default type", type_timeout_min=8 * 60)
            self.getMachineTypeRepository(session).create(mt1)

            r1 = Role(role_name="admins", authorize_all=True, reserved=True, maintenance=True, backend_admin=True)
            self.getRoleRepository(session).create(r1)

            r3 = Role(role_name="Fab Staff", authorize_all=True, reserved=False, maintenance=True, backend_admin=False)
            self.getRoleRepository(session).create(r3)

            r2 = Role(
                role_name="Fab Users", authorize_all=True, reserved=False, maintenance=False, backend_admin=False
            )
            self.getRoleRepository(session).create(r2)

            r4 = self.seedAnonymousRole(session)
            self.seedAnonymousUser(session, r4)

            # Create default admin user
            u1 = User(
                name="admin",
                surname="admin",
                role_id=r1.role_id,
                card_UUID="12345678",
                email=getSetting("web", "default_admin_email"),
            )
            u1.set_password(User.DEFAULT_ADMIN_PASSWORD)
            self.getUserRepository(session).create(u1)

            m1 = Machine(machine_name="MACHINE1", machine_type_id=mt1.type_id)
            self.getMachineRepository(session).create(m1)

    def seedAnonymousRole(self, session) -> Role:
        anrole = Role(
            role_name="Anonymous", authorize_all=False, reserved=True, maintenance=False, backend_admin=False
        )
        self.getRoleRepository(session).create(anrole)
        return anrole

    def seedAnonymousUser(self, session, role) -> User:
        anon = User(
            name="Anonymous",
            surname="",
            role_id=role.role_id,
            card_UUID=None,
            email="",
        )
        self.getUserRepository(session).create(anon)
        return anon

    def dropContents(self) -> None:
        """Drop all contents of the database."""
        logging.warning("Dropping all contents of database: %s", self._url)
        meta = Base.metadata
        with self._session() as session:
            trans = session.begin()
            for table in reversed(meta.sorted_tables):
                session.execute(table.delete())
            trans.commit()

    def copy(self, destination: str) -> None:
        """Copy the database to a new location.

        Args:
            destination (str): The destination of the copy.
        """
        from shutil import copyfile

        copyfile(self._url.replace("sqlite:///", ""), destination)
        logging.info("Copied database from %s to %s", self._name, destination)



from time import time
import unittest
from FabOMatic.database.models import Authorization, MachineType, Use

from FabOMatic.mqtt.mqtt_types import (
    UserQuery,
    AliveQuery,
    MachineQuery,
    StartUseQuery,
    EndUseQuery,
    RegisterMaintenanceQuery,
    InUseQuery,
)

from FabOMatic.mqtt.MQTTInterface import MQTTInterface
from FabOMatic.logic.MsgMapper import MsgMapper
from FabOMatic.logic.MachineLogic import MachineLogic
from tests.common import TEST_SETTINGS_PATH, get_simple_db


class TestLogic(unittest.TestCase):
    def test_missed_messages_1(self):
        db = get_simple_db()
        with db.getSession() as session:
            MachineLogic.database = db
            mac = db.getMachineRepository(session).get_all()[0]
            user = db.getUserRepository(session).get_all()[0]
            user.card_UUID = "1234"
            db.getUserRepository(session).update(user)
            ml = MachineLogic(mac.machine_id)

            use_repo = db.getUseRepository(session)

            for use in use_repo.get_all():
                use_repo.delete(use)

            # Call inUse without startUse
            response = ml.inUse("1234", 10)
            self.assertTrue(response.request_ok, "inUse succeeded")

            session.commit()
            # Check a usage has been created with the correct duration
            usage = session.query(Use).filter(Use.machine_id.__eq__(mac.machine_id)).first()
            self.assertIsNotNone(usage, "Usage not created")
            self.assertAlmostEqual(time() - usage.start_timestamp, 10, 0, "Usage duration is not correct")
            self.assertIsNone(usage.end_timestamp, "Usage end timestamp is not None (1)")

            # Call again with another duration
            response = ml.inUse("1234", 20)

            session.commit()
            # Check that no new usage has been created and start time is not updated
            usage = session.query(Use).filter(Use.machine_id.__eq__(mac.machine_id)).first()
            self.assertEqual(len(use_repo.get_all()), 1, "Usage has been duplicated")
            self.assertTrue(response.request_ok, "inUse (2) failed")
            self.assertAlmostEqual(time() - usage.start_timestamp, 10, 0, "Usage duration is not updated")
            self.assertIsNone(usage.end_timestamp, "Usage end timestamp is not None (2)")

            session.commit()

            # Finally close the record
            response = ml.endUse("1234", 30, True)  # duration will override the previous calculated start
            self.assertTrue(response.request_ok, "endUse failed")

            session.commit()
            usage = session.query(Use).filter(Use.machine_id.__eq__(mac.machine_id)).first()
            self.assertIsNotNone(usage.end_timestamp, f"Usage end timestamp is None : {response}")
            self.assertAlmostEqual(usage.end_timestamp - usage.start_timestamp, 30, 0, "Usage duration is not correct")
            self.assertEqual(len(use_repo.get_all()), 1, "Usage has been duplicated")

    def test_missed_messages_2(self):
        db = get_simple_db()
        with db.getSession() as session:
            MachineLogic.database = db

            user = db.getUserRepository(session).get_all()[0]
            user.card_UUID = "1234"
            db.getUserRepository(session).update(user)

            use_repo = db.getUseRepository(session)

            for mac_id in range(1, 4):
                mac = db.getMachineRepository(session).get_by_id(mac_id)
                ml = MachineLogic(mac.machine_id)

                for use in use_repo.get_all():
                    use_repo.delete(use)

                session.commit()

                for d in range(10):
                    # Call inUse without startUse
                    response = ml.inUse("1234", 10 + d)
                    self.assertTrue(response.request_ok, "inUse succeeded")

                    # Check a usage has been created with the correct duration
                    usage = session.query(Use).filter(Use.machine_id == mac.machine_id).first()
                    self.assertIsNotNone(usage, "Usage not created")
                    self.assertAlmostEqual(time() - usage.start_timestamp, 10, 0, "Usage duration is not correct")
                    self.assertIsNone(usage.end_timestamp, "Usage end timestamp is not None (1)")

                    # Check for duplicated records
                    self.assertEqual(len(use_repo.get_all()), 1, "Usage has been duplicated")

                # Call again with another duration
                response = ml.startUse("1234", False)
                self.assertTrue(response.request_ok, "startUse failed")

                session.commit()
                # Check that inuse has been closed with duration 10s
                usage = (
                    session.query(Use)
                    .filter(Use.machine_id == mac.machine_id)
                    .order_by(Use.start_timestamp.asc())
                    .first()
                )
                self.assertIsNotNone(usage.end_timestamp, f"Usage end timestamp is None : {usage}")
                self.assertAlmostEqual(
                    usage.end_timestamp - usage.start_timestamp, 10, 0, f"Usage duration is not correct {usage}"
                )

                # Check that another record has been opened
                usage = (
                    session.query(Use)
                    .filter(Use.machine_id == mac.machine_id)
                    .order_by(Use.start_timestamp.desc())
                    .first()
                )
                self.assertIsNone(usage.end_timestamp, "Usage end timestamp is not None (2)")

                response = ml.endUse("1234", 1, False)
                self.assertTrue(response.request_ok, "endUse failed")

                session.commit()
                usage = (
                    session.query(Use)
                    .filter(Use.machine_id == mac.machine_id)
                    .order_by(Use.start_timestamp.desc())
                    .first()
                )
                self.assertIsNotNone(usage.end_timestamp, f"Usage end timestamp is None : {response}")
                # Check that we have two records
                self.assertEqual(len(use_repo.get_all()), 2, "wrong number of request")

    def test_machine_logic(self):
        db = get_simple_db()
        with db.getSession() as session:
            MachineLogic.database = db
            mac = db.getMachineRepository(session).get_all()[0]
            user = db.getUserRepository(session).get_all()[0]

            user.card_UUID = "1234"
            db.getUserRepository(session).update(user)

            ml = MachineLogic(mac.machine_id)
            status = ml.machineStatus()
            self.assertTrue(status.is_valid, "Machine is not valid")
            self.assertTrue(status.request_ok, "Request is not ok")
            self.assertFalse(status.maintenance, "Machine is in maintenance")
            self.assertTrue(status.allowed, "Machine is not allowed")

            mac.blocked = True
            db.getMachineRepository(session).update(mac)
            status = ml.machineStatus()
            self.assertTrue(status.is_valid, "Machine is not valid")
            self.assertTrue(status.request_ok, "Request is not ok")
            self.assertFalse(status.maintenance, "Machine is in maintenance")
            self.assertFalse(status.allowed, "Machine is allowed")

            mac.blocked = False
            db.getMachineRepository(session).update(mac)

            query = AliveQuery("0.1.32", "127.0.0.1", "SN0", 1000)
            ml.machineAlive(query)

            response = ml.isAuthorized("1234")
            self.assertTrue(response.request_ok, "isAuthorized failed")
            self.assertTrue(response.is_valid, "isAuthorized returns invalid with valid card")

            response = ml.isAuthorized("12345678")
            self.assertTrue(response.request_ok, "isAuthorized must succeed with invalid card")
            self.assertFalse(response.is_valid, "isAuthorized must return invalid with invalid card")

            response = ml.startUse("1234", True)
            self.assertTrue(response.request_ok, "startUse failed")

            response = ml.inUse("1234", 1)
            self.assertTrue(response.request_ok, "inUse failed")

            response = ml.endUse("1234", 123, True)
            self.assertTrue(response.request_ok, "endUse failed")

            ml.registerMaintenance("1234", False)
            self.assertTrue(response.request_ok, "registerMaintenance failed")

            ml.registerMaintenance("1234", True)
            self.assertTrue(response.request_ok, "registerMaintenance failed")

    def test_msg_mapper(self):
        db = get_simple_db()

        mqtt = MQTTInterface(TEST_SETTINGS_PATH)
        mqtt.connect()
        mapper = MsgMapper(mqtt, db)
        mapper.registerHandlers()

        # Try all messagges
        query = UserQuery("1234")
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = AliveQuery("0.1.32", "127.0.0.1", "SN0", 1000)
        self.assertFalse(mapper.messageReceived("1", query), "Alive message has no response")
        query = MachineQuery()
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = StartUseQuery("1234", False)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = InUseQuery("1234", 123)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = EndUseQuery("1234", 123, False)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = RegisterMaintenanceQuery("1234", False)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")

        # Try all with invalid card
        query = UserQuery("DEADBEEF")
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = StartUseQuery("DEADBEEF", False)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = InUseQuery("DEADBEEF", 123)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = EndUseQuery("DEADBEEF", 123, False)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")
        query = RegisterMaintenanceQuery("DEADBEEF", True)
        self.assertTrue(mapper.messageReceived("1", query), "Message not processed")

    def test_machine_auth(self):
        db = get_simple_db()
        with db.getSession() as session:
            MachineLogic.database = db
            mac = db.getMachineRepository(session).get_all()[0]
            user = db.getUserRepository(session).get_all()[0]
            user.card_UUID = "1234"
            db.getUserRepository(session).update(user)

            ml = MachineLogic(mac.machine_id)
            status = ml.machineStatus()
            self.assertTrue(status.is_valid, "Machine is not valid")
            self.assertTrue(status.request_ok, "Request is not ok")
            self.assertFalse(status.maintenance, "Machine is in maintenance")
            self.assertTrue(status.allowed, "Machine is not allowed")

            mac.blocked = False
            db.getMachineRepository(session).update(mac)

            mac.machine_type.access_management = MachineType.MANAGEMENT_WITH_AUTHORIZATION
            db.getMachineTypeRepository(session).update(mac.machine_type)

            # Test success
            response = ml.isAuthorized("1234")
            self.assertTrue(response.request_ok, "isAuthorized failed")
            self.assertTrue(response.is_valid, "isAuthorized returns invalid with authorized user")

            try:
                authorization = (
                    session.query(Authorization).filter_by(user_id=user.user_id, machine_id=mac.machine_id).one()
                )
                session.delete(authorization)
                session.commit()
            except:
                pass

            user.role.authorize_all = False
            db.getRoleRepository(session).update(user.role)

            # Test failure
            response = ml.isAuthorized("1234")
            self.assertTrue(response.request_ok, "isAuthorized failed")
            self.assertFalse(response.is_valid, "isAuthorized returns true without authorization by machine type")

            # Authorize all
            mac.machine_type.access_management = MachineType.MANAGEMENT_WITHOUT_AUTHORIZATION
            db.getMachineTypeRepository(session).update(mac.machine_type)

            # Test success without authorization
            response = ml.isAuthorized("1234")
            self.assertTrue(response.request_ok, "isAuthorized failed")
            self.assertTrue(
                response.is_valid, "isAuthorized returns invalid while machine type has no authorization required"
            )



""" Routes for machines management. """

# pylint: disable=C0116

from flask import flash, render_template, request, redirect, url_for
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import Machine, MachineType
from FabOMatic.database.repositories import MachineRepository
from .webapplication import DBSession, app


@app.route("/machines", methods=["GET"])
@login_required
def view_machines():
    session = DBSession()
    machine_repo = MachineRepository(session)
    machines = session.query(Machine).order_by(Machine.machine_id).all()
    maint_stats = {}
    for machine in machines:
        for maint in machine.maintenances:
            elapsed_s = machine_repo.getRelativeUseTimeByMaintenance(machine.machine_id, maint)
            elapsed_h = round(elapsed_s / 3600.0, 1)
            expired = elapsed_h > maint.hours_between
            maint_stats[maint.maintenance_id] = {"expired": expired, "elapsed": elapsed_h}

    return render_template("view_machines.html", machines=machines, maint_stats=maint_stats)


@app.route("/machines/add", methods=["GET"])
@login_required
def add_machine():
    session = DBSession()
    machine_types = session.query(MachineType).order_by(MachineType.type_id).all()
    return render_template("add_machine.html", machine_types=machine_types)


@app.route("/machines/create", methods=["POST"])
@login_required
def create_machine():
    session = DBSession()
    machine_data = request.form

    # Input validation
    try:
        if float(machine_data["machine_hours"]) < 0:
            flash(gettext("Hours must be a positive number."))
            return redirect(url_for("add_machine"))
    except ValueError:
        flash(gettext("Hours must be a number."))
        return redirect(url_for("add_machine"))

    new_machine = Machine(
        machine_name=machine_data["machine_name"],
        machine_type_id=machine_data["machine_type_id"],
        machine_hours=float(machine_data["machine_hours"]),
        blocked=machine_data.get("blocked", "off") == "on",
    )
    session.add(new_machine)
    session.commit()
    return redirect(url_for("view_machines"))


@app.route("/machines/edit/<int:machine_id>", methods=["GET"])
@login_required
def edit_machine(machine_id):
    session = DBSession()
    machine = session.query(Machine).filter_by(machine_id=machine_id).one()
    machine_types = session.query(MachineType).order_by(MachineType.type_id).all()
    if machine:
        return render_template("edit_machine.html", machine=machine, machine_types=machine_types)
    else:
        return gettext("Machine not found"), 404


@app.route("/machines/update", methods=["POST"])
@login_required
def update_machine():
    session = DBSession()
    machine_data = request.form
    machine = session.query(Machine).filter_by(machine_id=machine_data["machine_id"]).one()
    if machine:
        machine.machine_name = machine_data["machine_name"]
        machine.machine_type_id = machine_data["machine_type_id"]

        # Input validation
        try:
            if float(machine_data["machine_hours"]) < 0:
                flash(gettext("Hours must be a positive number."))
                return redirect(url_for("edit_machine", machine_id=machine.machine_id))
        except ValueError:
            flash(gettext("Hours must be a number."))
            return redirect(url_for("edit_machine", machine_id=machine.machine_id))

        machine.machine_hours = float(machine_data["machine_hours"])
        machine.blocked = machine_data.get("blocked", "off") == "on"
        session.commit()
        return redirect(url_for("view_machines"))
    else:
        return gettext("Machine not found"), 404


@app.route("/machines/delete/<int:machine_id>", methods=["GET", "POST"])
@login_required
def delete_machine(machine_id):
    session = DBSession()
    machine = session.query(Machine).filter_by(machine_id=machine_id).one()
    if not machine:
        return gettext("Machine not found"), 404

    if request.method == "POST":
        session.delete(machine)
        session.commit()
        return redirect(url_for("view_machines"))

    return render_template("delete_machine.html", machine=machine)



""" Routes for Machine Types management """

# pylint: disable=C0116

from flask import render_template, request, redirect, url_for, flash
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import MachineType
from .webapplication import DBSession, app


@app.route("/machinetypes")
@login_required
def machinetypes():
    session = DBSession()
    machine_types = session.query(MachineType).all()
    return render_template("view_machinetypes.html", machine_types=machine_types)


@app.route("/machinetypes/add", methods=["GET", "POST"])
@login_required
def add_machinetype():
    if request.method == "POST":
        session = DBSession()
        type_name = request.form["type_name"]
        timeout_min = request.form["type_timeout_min"]
        grace_period_min = request.form["grace_period_min"]
        access_management = request.form["access_control"]

        if not timeout_min.isnumeric() or not grace_period_min.isnumeric() or not access_management.isnumeric():
            flash(gettext("Invalid values for timeout or grace period or access control."))
            return redirect(url_for("add_machinetype"))
        else:
            timeout_min = int(timeout_min)
            grace_period_min = int(grace_period_min)
            access_management = int(access_management)

        if timeout_min < 0 or timeout_min > 65535:
            flash(gettext("Invalid values for timeout"))
            return redirect(url_for("add_machinetype"))

        if grace_period_min < 0 or grace_period_min > 65535:
            flash(gettext("Invalid values for grace period."))
            return redirect(url_for("add_machinetype"))

        machine_type = MachineType(
            type_name=type_name,
            type_timeout_min=timeout_min,
            grace_period_min=grace_period_min,
            access_management=access_management,
        )
        session.add(machine_type)
        session.commit()
        return redirect(url_for("machinetypes"))
    else:
        return render_template("add_machinetype.html", MachineType=MachineType)


@app.route("/machinetypes/edit/<int:type_id>", methods=["GET", "POST"])
@login_required
def edit_machinetype(type_id):
    session = DBSession()
    machine_type = session.query(MachineType).filter_by(type_id=type_id).one()
    if request.method == "POST":
        grace_period_min = request.form["grace_period_min"]
        timeout_min = request.form["type_timeout_min"]
        access_management = request.form["access_control"]

        if not timeout_min.isnumeric() or not grace_period_min.isnumeric() or not access_management.isnumeric():
            flash(gettext("Invalid values for timeout or grace period or access control."))
            return redirect(url_for("edit_machinetype", type_id=type_id))
        else:
            timeout_min = int(timeout_min)
            grace_period_min = int(grace_period_min)
            access_management = int(access_management)

        machine_type.type_timeout_min = timeout_min
        machine_type.grace_period_min = grace_period_min
        machine_type.access_management = access_management
        machine_type.type_name = request.form["type_name"]

        if machine_type.type_timeout_min < 0 or machine_type.type_timeout_min > 65535:
            flash(gettext("Invalid values for timeout"))
            return redirect(url_for("edit_machinetype", type_id=type_id))

        if machine_type.type_timeout_min < 0 or machine_type.grace_period_min > 65535:
            flash(gettext("Invalid values for grace period."))
            return redirect(url_for("edit_machinetype", type_id=type_id))

        session.add(machine_type)
        session.commit()
        return redirect(url_for("machinetypes"))
    else:
        return render_template("edit_machinetype.html", type_id=machine_type.type_id, machine_type=machine_type)


@app.route("/machinetypes/delete/<int:type_id>", methods=["GET", "POST"])
@login_required
def delete_machinetype(type_id):
    session = DBSession()
    machine_type = session.query(MachineType).filter_by(type_id=type_id).one()
    if not machine_type:
        return gettext("Machine Type not found"), 404
    if request.method == "POST":
        session.delete(machine_type)
        session.commit()
        flash(gettext("Machine type deleted successfully."))
        return redirect(url_for("machinetypes"))
    return render_template("delete_machinetype.html", machine_type=machine_type)



"""Maintenance: added URL field and removed attachments.

Revision ID: aa9ed3e094d5
Revises: 3cb12dfd8d6e
Create Date: 2024-04-24 22:39:03.511051

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "aa9ed3e094d5"
down_revision: Union[str, None] = "3cb12dfd8d6e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

from alembic import op
from sqlalchemy import engine_from_config
from sqlalchemy import inspect


def _table_has_column(table, column):
    config = op.get_context().config
    engine = engine_from_config(config.get_section(config.config_ini_section), prefix="sqlalchemy.")
    insp = inspect(engine)
    has_column = False
    for col in insp.get_columns(table):
        if column not in col["name"]:
            continue
        has_column = True
    return has_column


def _is_sqllite():
    return op.get_context().dialect.name == "sqlite"


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    if not _table_has_column("maintenances", "instructions_url"):
        op.add_column("maintenances", sa.Column("instructions_url", sa.String(), nullable=True))
    if _table_has_column("maintenances", "attachment"):
        if not _is_sqllite():
            op.drop_column("maintenances", "attachment")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    if _table_has_column("maintenances", "instructions_url"):
        if not _is_sqllite():
            op.drop_column("maintenances", "instructions_url")

    if not _table_has_column("maintenances", "attachment"):
        op.add_column("maintenances", sa.Column("attachment", sa.String(), nullable=True))
    # ### end Alembic commands ###



"""Database models for the FabOMatic application."""

import sqlite3
import os
from itsdangerous import URLSafeTimedSerializer as Serializer
from sqlalchemy import Column, Integer, String, Float, Boolean, ForeignKey, UniqueConstraint
from sqlalchemy import event, Engine, Index
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship
from werkzeug.security import generate_password_hash, check_password_hash
from flask_login import UserMixin

from .constants import DEFAULT_TIMEOUT_MINUTES, USER_LEVEL

Base = declarative_base()


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """Set SQLite PRAGMA foreign_keys=ON."""
    if isinstance(dbapi_connection, sqlite3.Connection):  # play well with other DB backends
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.execute("pragma auto_vacuum = incremental;")
        cursor.execute("pragma optimize;")
        cursor.close()


class Role(Base):
    """Dataclass handling a role."""

    __tablename__ = "roles"

    role_id = Column(Integer, primary_key=True, autoincrement=True)
    role_name = Column(String, unique=True, nullable=False)
    authorize_all = Column(Boolean, default=False, nullable=False)
    reserved = Column(Boolean, default=False, nullable=False)
    maintenance = Column(Boolean, default=False, nullable=False)
    backend_admin = Column(Boolean, default=False, nullable=False)

    users = relationship("User", back_populates="role")
    __table_args__ = (Index("idx_roles_role_name_unique", "role_name", unique=True),)

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "role_id": self.role_id,
            "role_name": self.role_name,
            "authorize_all": self.authorize_all,
            "reserved": self.reserved,
            "maintenance": self.maintenance,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)


class User(UserMixin, Base):
    """Dataclass handling a user."""

    __tablename__ = "users"
    DEFAULT_ADMIN_PASSWORD = "admin"

    user_id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    surname = Column(String, nullable=False)
    role_id = Column(Integer, ForeignKey("roles.role_id"), nullable=False)
    card_UUID = Column(String, unique=True, nullable=True)
    disabled = Column(Boolean, unique=False, nullable=False, default=False)
    deleted = Column(Boolean, unique=False, nullable=False, default=False)
    password_hash = Column(String(128), nullable=True)
    email = Column(String, nullable=True)

    authorizations = relationship("Authorization", back_populates="user")
    interventions = relationship("Intervention", back_populates="user")
    uses = relationship("Use", back_populates="user")
    role = relationship("Role", back_populates="users")

    __table_args__ = (Index("idx_users_card_UUID_unique", "card_UUID", unique=True),)

    """ Methods required by Flask-Login """

    def set_password(self, password) -> None:
        self.password_hash = generate_password_hash(password)

    def check_password(self, password) -> bool:
        return check_password_hash(self.password_hash, password)

    def get_id(self) -> str:
        return str(self.user_id)

    """ Email reset """

    def get_reset_token(self, key: bytes, salt: bytes) -> str:
        s = Serializer(secret_key=key, salt=salt)
        return s.dumps({"user_id": self.user_id})

    """ Static method to verify a token"""
    """ Returns the user_id if the token is valid, None otherwise """
    """ max_age is in seconds"""
    """ key and salt must be the same used to generate the token """

    @staticmethod
    def verify_reset_token(token: str, key: bytes, salt: bytes, max_age=1800) -> int | None:
        s = Serializer(secret_key=key, salt=salt)
        try:
            user_id = s.loads(s=token, max_age=max_age)["user_id"]
        except:
            return None

        return int(user_id)

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "user_id": self.user_id,
            "name": self.name,
            "surname": self.surname,
            "role_id": self.role_id,
            "authorization_ids": [a.authorization_id for a in self.authorizations],
            "intervention_ids": [i.intervention_id for i in self.interventions],
            "use_ids": [u.use_id for u in self.uses],
            "card_UUID": self.card_UUID,
            "disabled": self.disabled,
            "deleted": self.deleted,
            "email": self.email,
            "password_hash": self.password_hash,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)

    def user_level(self) -> USER_LEVEL:
        """Get the user level."""
        if self.disabled:
            return USER_LEVEL.INVALID
        if self.role.authorize_all:
            return USER_LEVEL.ADMIN
        return USER_LEVEL.NORMAL


class Authorization(Base):
    """Dataclass handling an authorization."""

    __tablename__ = "authorizations"

    authorization_id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey("users.user_id"), nullable=False)
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)

    user = relationship("User", back_populates="authorizations")
    machine = relationship("Machine", back_populates="authorizations")

    __table_args__ = (UniqueConstraint("user_id", "machine_id", name="uix_1"),)

    def serialize(self):
        """Serialize data and return a Dict."""
        return {"authorization_id": self.authorization_id, "user_id": self.user_id, "machine_id": self.machine_id}

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)


class Maintenance(Base):
    """Dataclass handling a maintenance."""

    __tablename__ = "maintenances"

    maintenance_id = Column(Integer, primary_key=True, autoincrement=True)
    hours_between = Column(Float, nullable=False)
    description = Column(String, nullable=False)
    lcd_message = Column(String, nullable=True, default="")
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)
    instructions_url = Column(String, nullable=True, default="")

    machine = relationship("Machine", back_populates="maintenances")
    interventions = relationship("Intervention", back_populates="maintenance")

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "maintenance_id": self.maintenance_id,
            "hours_between": self.hours_between,
            "description": self.description,
            "lcd_message": self.lcd_message,
            "machine_id": self.machine_id,
            "instructions_url": self.instructions_url,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)


class Intervention(Base):
    """Class handling an intervention."""

    __tablename__ = "interventions"

    intervention_id = Column(Integer, primary_key=True, autoincrement=True)
    maintenance_id = Column(Integer, ForeignKey("maintenances.maintenance_id"), nullable=False)
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)
    user_id = Column(Integer, ForeignKey("users.user_id"), nullable=False)
    timestamp = Column(Float, nullable=False)
    replay = Column(Boolean, nullable=True)

    machine = relationship("Machine", back_populates="interventions")
    maintenance = relationship("Maintenance", back_populates="interventions")
    user = relationship("User", back_populates="interventions")

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "intervention_id": self.intervention_id,
            "maintenance_id": self.maintenance_id,
            "machine_id": self.machine_id,
            "user_id": self.user_id,
            "timestamp": self.timestamp,
            "replay": self.replay,
        }


class MachineType(Base):
    """Dataclass handling a machine type."""

    __tablename__ = "machine_types"
    MANAGEMENT_WITH_AUTHORIZATION = 0
    MANAGEMENT_WITHOUT_AUTHORIZATION = 1

    type_id = Column(Integer, primary_key=True, autoincrement=True)
    type_name = Column(String, unique=True, nullable=False)
    type_timeout_min = Column(Integer, unique=False, nullable=False, default=DEFAULT_TIMEOUT_MINUTES)
    grace_period_min = Column(Integer, unique=False, nullable=True, default=2)
    access_management = Column(Integer, unique=False, nullable=True, default=MANAGEMENT_WITH_AUTHORIZATION)
    machines = relationship("Machine", back_populates="machine_type")
    __table_args__ = (Index("idx_machine_types_type_name_unique", "type_name", unique=True),)

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "type_id": self.type_id,
            "type_name": self.type_name,
            "type_timeout_min": self.type_timeout_min,
            "grace_period_min": self.grace_period_min,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)


class Machine(Base):
    """Class handling a machine."""

    __tablename__ = "machines"

    machine_id = Column(Integer, primary_key=True, autoincrement=True)
    machine_name = Column(String, unique=True, nullable=False)
    machine_type_id = Column(Integer, ForeignKey("machine_types.type_id"))
    machine_hours = Column(Float, nullable=False, default=0.0)  # Somewhat redundant with sum of uses duration
    blocked = Column(Boolean, nullable=False, default=False)
    last_seen = Column(Float, nullable=True)
    maintenances = relationship("Maintenance", back_populates="machine")
    interventions = relationship("Intervention", back_populates="machine")
    authorizations = relationship("Authorization", back_populates="machine")
    machine_type = relationship("MachineType", back_populates="machines", lazy=False)
    uses = relationship("Use", back_populates="machine")
    cards = relationship("UnknownCard", back_populates="machine")
    boards = relationship("Board", back_populates="machine")

    __table_args__ = (Index("idx_machines_machine_name_unique", "machine_name", unique=True),)

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "machine_id": self.machine_id,
            "machine_name": self.machine_name,
            "machine_type_id": self.machine_type_id,
            "machine_hours": self.machine_hours,
            "blocked": self.blocked,
            "maintenances": [maintenance.maintenance_id for maintenance in self.maintenances],
            "interventions": [intervention.intervention_id for intervention in self.interventions],
        }

    def active_user(self) -> User | None:
        """Get the current user."""
        for use in reversed(self.uses):
            if use.end_timestamp is None:
                return use.user
        return None

    def active_board(self):
        """Get the current board."""
        if len(self.boards) > 0:
            return sorted(self.boards, key=lambda b: b.last_seen, reverse=True)[0]
        return None


class Use(Base):
    """Class handling machine use."""

    __tablename__ = "uses"

    use_id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey("users.user_id"), nullable=False)
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)
    start_timestamp = Column(Float, nullable=False)
    last_seen = Column(Float, nullable=False)
    end_timestamp = Column(Float, nullable=True)
    replay = Column(Boolean, nullable=True)

    machine = relationship("Machine", back_populates="uses")
    user = relationship("User", back_populates="uses")

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "use_id": self.use_id,
            "user_id": self.user_id,
            "machine_id": self.machine_id,
            "start_timestamp": self.start_timestamp,
            "end_timestamp": self.end_timestamp,
            "last_seen": self.last_seen,
            "replay": self.replay,
        }

    def __str__(self):
        """Return a string representation of the object."""
        return str(self.serialize())


class UnknownCard(Base):
    """Class handling machine use."""

    __tablename__ = "unknown_cards"
    id = Column(Integer, primary_key=True, autoincrement=True)
    card_UUID = Column(String, nullable=False)
    timestamp = Column(Float, nullable=False)
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)

    machine = relationship("Machine", back_populates="cards")

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "id": self.id,
            "card_UUID": self.card_UUID,
            "timestamp": self.timestamp,
            "machine_id": self.machine_id,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)


class Board(Base):
    """Dataclass handling a role."""

    __tablename__ = "boards"

    board_id = Column(Integer, primary_key=True, autoincrement=True)
    machine_id = Column(Integer, ForeignKey("machines.machine_id"), nullable=False)
    ip_address = Column(String, unique=False, nullable=False)
    fw_version = Column(String, unique=False, nullable=False)
    serial = Column(String, unique=False, nullable=True, default="?")
    heap = Column(Integer, unique=False, nullable=True, default=0)

    last_seen = Column(Float, nullable=False)

    machine = relationship("Machine", back_populates="boards")

    def serialize(self):
        """Serialize data and return a Dict."""
        return {
            "board_id": self.board_id,
            "machine_id": self.machine_id,
            "ip_address": self.ip_address,
            "fw_version": self.fw_version,
            "serial": self.serial,
            "heap": self.heap,
            "last_seen": self.last_seen,
        }

    @classmethod
    def from_dict(cls, dict_data):
        """Deserialize data from Dictionary."""
        return cls(**dict_data)



"""Maintenance: added short lcd message

Revision ID: 3cb12dfd8d6e
Revises: e544554c7d29
Create Date: 2024-04-24 22:17:20.179783

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '3cb12dfd8d6e'
down_revision: Union[str, None] = 'e544554c7d29'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('maintenances', sa.Column('lcd_message', sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('maintenances', 'lcd_message')
    # ### end Alembic commands ###



""" Routes for managing roles. """

# pylint: disable=C0116

from flask import flash, redirect, render_template, Response, request, send_file, url_for
from flask_login import login_required
from flask_babel import gettext
from importlib.metadata import version

from FabOMatic.database.DatabaseBackend import getDatabaseUrl
from FabOMatic.database.repositories import BoardsRepository
from .webapplication import DBSession, app
import os
import platform
import shutil
import psutil
import subprocess
import requests


@app.route("/system")
@login_required
def system():
    db_file = getDatabaseUrl().replace("sqlite:///", "")
    db_size = os.path.getsize(db_file)
    # Returns information about the host machine (Raspberry Pi)
    machine_info = f"{platform.uname() }, CPU:{os.cpu_count()}"
    total, used, disk_free = shutil.disk_usage(__file__)
    stats = psutil.virtual_memory()
    ram_free = getattr(stats, "available")

    # application details
    app_version = version("FabOMatic")

    # check if there is an updated pypi package
    package = "FabOMatic"  # replace with the package you want to check
    response = requests.get(f"https://pypi.org/pypi/{package}/json")
    latest_version = response.json()["info"]["version"]

    # Get the boards from the database repository
    session = DBSession()
    board_repo = BoardsRepository(session)
    boards = board_repo.get_all()

    return render_template(
        "view_system.html",
        db_size=int(db_size / 1024),
        machine_info=machine_info,
        disk_free=int(disk_free / 1024 / 1024),
        ram_free=int(ram_free / 1024 / 1024),
        app_version=app_version,
        latest_version=latest_version,
        db_file=db_file,
        boards=boards,
    )


@app.route("/download_db")
@login_required
def download_db():
    # Returns of copy of the SQLite database to the user
    db_file = getDatabaseUrl().replace("sqlite:///", "")
    return send_file(db_file, as_attachment=True)


@app.route("/download_logs")
@login_required
def download_logs():
    log_dir = os.path.expanduser("~/log")
    log_file = os.path.join(log_dir, "log.txt")
    return send_file(log_file, as_attachment=True)


@app.route("/upload_db", methods=["POST"])
@login_required
def upload_db():
    if "db_file" not in request.files:
        return redirect(request.url)

    new_db_file = request.files["db_file"]
    if not new_db_file.filename.lower().endswith(".sqldb"):
        flash(gettext("File must have sqldb extension"), "error")
        redirect(url_for("system"))

    # Backup existing file
    actual_db_file = getDatabaseUrl().replace("sqlite:///", "")
    backup_copy = actual_db_file + ".bak"
    shutil.copyfile(actual_db_file, backup_copy)

    # Save the uploaded file over the existing
    new_db_file.save(actual_db_file)
    # Redirect to the home page after uploading
    flash(gettext("Database was replaced. Previous copy can be found at ") + backup_copy)
    return redirect(url_for("system"))


@app.route("/update_app")
@login_required
def update_app():
    upgrade_output = subprocess.run(
        ["pip", "install", "-i", "https://pypi.org/simple/", "FabOMatic", "--upgrade"],
        check=True,
        text=True,
        capture_output=True,
    )

    # Restart the application using Systemd
    if platform.system() != "Windows":
        restart_output = subprocess.run(
            ["systemctl", "--user", "restart", "fablab"], check=True, text=True, capture_output=True
        )
        restart_output = f"Restart output:\n{restart_output.stdout}"
    else:
        restart_output = "Restart skipped because application is running on Windows or not running under systemd."

    output = f"Upgrade results : PIP output:\n{upgrade_output}\n\nSystemd output:\n{restart_output}"

    return Response(output, mimetype="text/plain")


@app.route("/reboot")
@login_required
def reboot():
    if platform.system() != "Windows":
        reboot_output = subprocess.run(["sudo", "reboot"], check=True, text=True, capture_output=True)
        reboot_output = f"Reboot output:\n{reboot_output.stdout}"
    else:
        reboot_output = "Reboot skipped because application is running on Windows"

    return Response(reboot_output, mimetype="text/plain")


@app.route("/restart_app")
@login_required
def restart_app():
    # Restart the application using Systemd
    if platform.system() != "Windows":
        restart_output = subprocess.run(
            ["systemctl", "--user", "restart", "fablab"], check=True, text=True, capture_output=True
        )
        restart_output = f"Restart output:\n{restart_output.stdout}"
    else:
        restart_output = "Restart skipped because application is running on Windows or not running under systemd."

    return Response(restart_output, mimetype="text/plain")



"""Boards: added columns for serial number and free heap

Revision ID: e544554c7d29
Revises: 123b2cca325f
Create Date: 2024-04-21 13:24:54.520194

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'e544554c7d29'
down_revision: Union[str, None] = '123b2cca325f'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('boards', sa.Column('serial', sa.String(), nullable=True))
    op.add_column('boards', sa.Column('heap', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('boards', 'heap')
    op.drop_column('boards', 'serial')
    # ### end Alembic commands ###



""" Test with more data. """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import unittest

from FabOMatic.__main__ import Backend
from tests.common import TEST_SETTINGS_PATH, add_test_data, get_simple_db


class TestData(unittest.TestCase):
    def test_data(self):
        db = get_simple_db()
        add_test_data(db, 2000)



""" This module contains common functions used in the tests. """

import logging
import os
from logging.handlers import RotatingFileHandler
import random
from time import time

from sqlalchemy import text

from FabOMatic.database.DatabaseBackend import DatabaseBackend, getSetting
from FabOMatic.database.models import Machine, MachineType, Maintenance, Role, Use, User, Intervention

FIXTURE_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_SETTINGS_PATH = os.path.join(FIXTURE_DIR, "test_settings.toml")


def configure_logger():
    """
    Configures a logger object with a rotating file handler.

    The logger is named "test_logger" and is set to log messages at the DEBUG level.
    The logs are formatted with the timestamp, log level, and message.
    The logs are written to a file named "test-log.txt" in the same directory as this script.
    The file handler rotates the log file when it reaches a maximum size of 1 MB, keeping 1 backup file.
    The file is encoded using the "latin-1" encoding.

    Returns:
        None
    """
    logger = logging.getLogger("test_logger")
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")

    log_file = os.path.join(os.path.dirname(__file__), "test-log.txt")
    file_handler = RotatingFileHandler(log_file, maxBytes=1000000, backupCount=1, encoding="latin-1")
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)


def get_empty_test_db() -> DatabaseBackend:
    """
    Returns an instance of DatabaseBackend with an empty database.

    Returns:
        DatabaseBackend: An instance of DatabaseBackend with an empty database.
    """
    d = DatabaseBackend(TEST_SETTINGS_PATH)
    d.deleteExistingDatabase()
    d.createAndUpdateDatabase()
    d.dropContents()
    return d


def seed_test_db() -> DatabaseBackend:
    """
    Seeds the database with initial data for testing purposes.

    Returns:
        DatabaseBackend: The seeded database instance.
    """
    empty_db = DatabaseBackend(TEST_SETTINGS_PATH)
    empty_db.deleteExistingDatabase()
    empty_db.createAndUpdateDatabase()
    empty_db.dropContents()

    with empty_db.getSession() as session:
        mt1 = MachineType(type_id=1, type_name="Default type")
        empty_db.getMachineTypeRepository(session).create(mt1)

        r1 = Role(role_name="admins", authorize_all=True, reserved=True, maintenance=True, backend_admin=True)
        empty_db.getRoleRepository(session).create(r1)

        r3 = Role(role_name="Fab Staff", authorize_all=False, reserved=False, maintenance=True, backend_admin=False)
        empty_db.getRoleRepository(session).create(r3)

        r2 = Role(role_name="Fab Users", authorize_all=False, reserved=False, maintenance=False, backend_admin=False)
        empty_db.getRoleRepository(session).create(r2)

        u1 = User(
            name="admin",
            surname="admin",
            role_id=r1.role_id,
            card_UUID="12345678",
            email=getSetting("web", "default_admin_email"),
        )

        u1.set_password(User.DEFAULT_ADMIN_PASSWORD)

        empty_db.getUserRepository(session).create(u1)

        m1 = Machine(machine_name="Sample machine", machine_type_id=mt1.type_id)
        empty_db.getMachineRepository(session).create(m1)

        maint1 = Maintenance(
            hours_between=10,
            description="sample maintenance - clean machine",
            machine_id=m1.machine_id,
            lcd_message="allineare",
            instructions_url="https://www.fablabbergamo.it/",
        )
        empty_db.getMaintenanceRepository(session).create(maint1)

    return empty_db


def add_test_data(db: DatabaseBackend, nb_records: int) -> DatabaseBackend:
    with db.getSession() as session:
        use_repo = db.getUseRepository(session)
        inter_repo = db.getInterventionRepository(session)
        # generate 100 more users
        users = []
        for i in range(100, 201):
            temp_user = User(
                name="User" + str(i),
                surname="Surname" + str(i),
                role_id=2,
                card_UUID=(str(i * 100) * 8)[:8],
                disabled=random.choice([True, False]),
            )
            temp_user.set_password("")
            temp_user.email = f"user{str(i)*2}@fablab.org"
            users.append(temp_user)
        db.getUserRepository(session).bulk_create(users)
        session.commit()

        for m in db.getMachineRepository(session).get_all():
            uses = []
            for i in range(1, nb_records):
                start = time() - 100000 * i
                end = start + random.randint(1, 100000)

                use = Use(
                    user_id=random.choice(range(1, 10)),
                    machine_id=m.machine_id,
                    start_timestamp=start,
                    end_timestamp=end,
                    last_seen=end,
                )
                uses.append(use)
            use_repo.bulk_create(uses)

            for mt in m.maintenances:
                interventions = []
                for i in range(1, nb_records // 5):
                    inter = Intervention(
                        maintenance_id=mt.maintenance_id,
                        user_id=random.choice(range(1, 10)),
                        machine_id=m.machine_id,
                        timestamp=time() - 100000 * i,
                    )
                    inter_repo.create(inter)
                inter_repo.bulk_create(interventions)
        session.commit()
    # Make a copy in current folder to ease testing with UI/archival into tests/databases
    db.copy("test-full.sqldb")
    return db


def get_simple_db() -> DatabaseBackend:
    """
    Creates a simple database with predefined data for testing purposes.

    Returns:
        DatabaseBackend: An instance of the created database.
    """
    empty_db = DatabaseBackend(TEST_SETTINGS_PATH)
    empty_db.deleteExistingDatabase()
    empty_db.createAndUpdateDatabase()
    empty_db.dropContents()

    with empty_db.getSession() as session:
        mt1 = MachineType(type_id=1, type_name="LASER")
        empty_db.getMachineTypeRepository(session).create(mt1)

        mt2 = MachineType(type_id=2, type_name="3D PRINTER")
        empty_db.getMachineTypeRepository(session).create(mt2)

        mt3 = MachineType(type_id=3, type_name="DRILL")
        empty_db.getMachineTypeRepository(session).create(mt3)

        r1 = Role(role_name="admin", authorize_all=True, reserved=True, maintenance=True)
        empty_db.getRoleRepository(session).create(r1)

        r3 = Role(role_name="staff", authorize_all=False, reserved=False, maintenance=True)
        empty_db.getRoleRepository(session).create(r3)

        r2 = Role(role_name="fab users", authorize_all=False, reserved=False, maintenance=False)
        empty_db.getRoleRepository(session).create(r2)

        u1 = User(name="Mario", surname="Rossi", role_id=r1.role_id, email="marco.rossi@fablab.org", card_UUID="1234")
        u1.set_password("password1")
        empty_db.getUserRepository(session).create(u1)

        u2 = User(
            name="Andrea", surname="Bianchi", role_id=r2.role_id, email="andrea.bianchi@fablab.org", card_UUID="5678"
        )
        u2.set_password("password2")
        empty_db.getUserRepository(session).create(u2)

        for i in range(1, 10):
            temp_user = User(
                name="User" + str(i),
                surname="Surname" + str(i),
                role_id=r2.role_id,
                card_UUID=(str(i) * 8)[:8],
                disabled=random.choice([True, False]),
            )
            temp_user.set_password("")
            temp_user.email = f"{temp_user.name}.{temp_user.surname}@fablab.org"
            empty_db.getUserRepository(session).create(temp_user)

        m1 = Machine(machine_name="LASER 1", machine_type_id=mt1.type_id)
        empty_db.getMachineRepository(session).create(m1)

        m2 = Machine(machine_name="PRINTER 1", machine_type_id=mt2.type_id)
        empty_db.getMachineRepository(session).create(m2)

        temp_machines = []
        for i in range(1, 10):
            temp_machine = Machine(machine_name="Machine" + str(i), machine_type_id=mt3.type_id)
            temp_machines.append(temp_machine)
            empty_db.getMachineRepository(session).create(temp_machine)

        maint1 = Maintenance(
            hours_between=10,
            description="replace engine",
            machine_id=m1.machine_id,
            lcd_message="sostituire motore",
            instructions_url="https://www.fablabbergamo.it/",
        )
        empty_db.getMaintenanceRepository(session).create(maint1)

        maint2 = Maintenance(
            hours_between=10,
            description="replace brushes",
            machine_id=m2.machine_id,
            lcd_message="pulire lente",
            instructions_url="https://www.fablabbergamo.it/",
        )
        empty_db.getMaintenanceRepository(session).create(maint2)

        for i in range(1, 10):
            temp_maint = Maintenance(
                hours_between=random.choice(range(1, 30)),
                description="Maintenance" + str(i),
                lcd_message="Message" + str(i),
                instructions_url="https://www.fablabbergamo.it/",
                machine_id=random.choice(temp_machines).machine_id,
            )
            empty_db.getMaintenanceRepository(session).create(temp_maint)

        timestamp = time() - 1000
        inter = Intervention(
            maintenance_id=maint1.maintenance_id, user_id=u1.user_id, machine_id=m1.machine_id, timestamp=timestamp
        )
        empty_db.getInterventionRepository(session).create(inter)

        inter2 = Intervention(
            maintenance_id=maint2.maintenance_id, user_id=u2.user_id, machine_id=m2.machine_id, timestamp=timestamp
        )
        empty_db.getInterventionRepository(session).create(inter2)

        session.commit()
    # Make a copy in current folder to ease testing with UI/archival into tests/databases
    empty_db.copy("test-simple.sqldb")
    return empty_db



""" Test the database backend. """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import random
import unittest
from random import randint
from string import ascii_uppercase
from time import time

from sqlalchemy.exc import IntegrityError

from FabOMatic.database.DatabaseBackend import DatabaseBackend
from FabOMatic.database.models import (
    Role,
    MachineType,
    User,
    Machine,
    Maintenance,
    Authorization,
    Use,
    Intervention,
)
from tests.common import TEST_SETTINGS_PATH, get_empty_test_db, get_simple_db, configure_logger


def random_string(length=16):
    """Generate a random string of fixed length"""
    return "".join(random.choice(ascii_uppercase) for i in range(length))


def random_between(a, b):
    return random.randint(a, b - 1)


class TestDB(unittest.TestCase):
    def test_connection(self):
        _ = DatabaseBackend(TEST_SETTINGS_PATH)

    def test_drop(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            self.assertEqual(len(empty_db.getMachineRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getMachineTypeRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getMachineRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getUserRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getRoleRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getMaintenanceRepository(session).get_all()), 0)
            self.assertEqual(len(empty_db.getInterventionRepository(session).get_all()), 0)

    def test_simple_add_roles(self):
        empty_db = get_empty_test_db()
        role_names = ["user", "power user", "moderator", "crew", "admin", "super admin"]

        with empty_db.getSession() as session:
            role_repo = empty_db.getRoleRepository(session)
            for i, r in enumerate(role_names):
                role_repo.create(
                    Role(
                        role_id=i, role_name=r, reserved=True, authorize_all=True, maintenance=True, backend_admin=True
                    )
                )

            # check if roles were added
            self.assertEqual(len(role_names), len(role_repo.get_all()))

            # check if role parameters are real
            for i, r in enumerate(role_repo.get_all()):
                self.assertEqual(r.role_id, i)
                self.assertEqual(r.role_name, role_names[i])
                self.assertTrue(r.reserved)
                self.assertTrue(r.maintenance)
                self.assertTrue(r.authorize_all)
                self.assertTrue(r.backend_admin)

            # edit a role name
            new_name = "TEST ROLE WOW"
            role = role_repo.get_by_id(0)
            role.role_name = new_name
            role.maintenance = False
            role_repo.update(role)
            self.assertEqual(role_repo.get_by_id(0).role_name, new_name)
            self.assertFalse(role_repo.get_by_id(0).maintenance)
            self.assertEqual(len(role_repo.get_all()), len(role_names))

            # create a new role
            new_role = "ÜBER ADMIN"
            role_repo.create(Role(role_name=new_role, backend_admin=True))
            self.assertEqual(len(role_repo.get_all()), len(role_names) + 1)
            self.assertTrue(role_repo.get_by_role_name(new_role).backend_admin)
            self.assertFalse(role_repo.get_by_role_name(new_role).maintenance)
            self.assertFalse(role_repo.get_by_role_name(new_role).authorize_all)
            self.assertFalse(role_repo.get_by_role_name(new_role).reserved)

            # Check autoincrement
            self.assertEqual(role_repo.get_by_id(len(role_names)).role_name, new_role)

            # remove a role
            role_repo.delete(role_repo.get_by_id(0))
            self.assertEqual(len(role_repo.get_all()), len(role_names))

            # add a new role
            role_repo.create(Role(role_name="ÜBERÜBER ADMIN"))
            self.assertEqual(len(role_repo.get_all()), len(role_names) + 1)

            # add a duplicate role and catch exception
            with self.assertRaises(IntegrityError):
                role_repo.create(Role(role_name="ÜBERÜBER ADMIN", role_id=1))

    def test_types(self):
        empty_db = get_empty_test_db()
        type_names = ["3d printer", "laser cutter", "vertical drill", "saw"]

        with empty_db.getSession() as session:
            type_repo = empty_db.getMachineTypeRepository(session)
            for i, t in enumerate(type_names):
                type_repo.create(MachineType(type_id=i, type_name=t))

            # check if types were added
            self.assertEqual(len(type_repo.get_all()), len(type_names))

            # rename a machine type
            new_name = "TEST TYPE WOW"
            mt0 = type_repo.get_model_by_id(MachineType, 0)
            mt0.type_name = new_name
            type_repo.update(mt0)

            self.assertEqual(type_repo.get_model_by_id(MachineType, 0).type_name, new_name)

            # add a new type
            new_type = "ÜBER TYPE"
            mt_x = MachineType(type_name=new_type)
            type_repo.create(mt_x)

            self.assertEqual(len(type_repo.get_all()), len(type_names) + 1)

            # remove a type
            type_repo.delete(mt_x)
            self.assertEqual(len(type_repo.get_all()), len(type_names))

            # list the types
            for t in type_repo.get_all():
                print(t.serialize())

            # add a duplicate type_name and catch exception
            a = MachineType(type_name="TEST TYPE WOW")
            with self.assertRaises(IntegrityError):
                type_repo.create(a)

            type_repo.rollback()

            a = MachineType(type_name="Unique name", type_id=1)
            with self.assertRaises(IntegrityError):
                type_repo.create(a)
            type_repo.rollback()

    def test_users(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            empty_db.getRoleRepository(session).create(Role(role_id=1, role_name="admin"))

            names = ["Alessandro", "Lorenzo", "Diego", "Tommaso", "Riccardo"]
            surnames = [
                "Rossi",
                "Bianchi",
                "Verdi",
                "Colombo",
                "Fumagalli",
            ]

            # just to make sure
            self.assertEqual(len(names), len(surnames))

            ids = []
            user_repo = empty_db.getUserRepository(session)
            # add roles
            for n, s in zip(names, surnames):
                usr = User(name=n, surname=s, role_id=1)
                user_repo.create(usr)
                self.assertEqual(user_repo.get_by_id(usr.user_id).name, n)
                self.assertEqual(user_repo.get_by_id(usr.user_id).surname, s)
                self.assertEqual(user_repo.get_by_id(usr.user_id).role_id, 1)
                ids.append(user_repo.get_by_id(usr.user_id).user_id)

            # check if user were added
            self.assertEqual(len(user_repo.get_all()), len(names))

            # check that every user is unique
            self.assertEqual(len(ids), len(set(ids)))

            # set a user card
            u = user_repo.get_by_id(ids[0])
            UUID = random_string(8)
            u.card_UUID = UUID
            user_repo.update(u)
            self.assertEqual(user_repo.get_by_id(ids[0]).card_UUID, UUID)

            # add another user
            user_repo.create(User(name="Andrea", surname="Bianchi", role_id=1))

            # test invalid role
            with self.assertRaises(IntegrityError):
                user_repo.create(User(name="Giorgio", surname="Rossi", role_id=10))

            user_repo.rollback()

            self.assertIsNone(user_repo.get_by_id(10))
            self.assertIsNone(user_repo.getUserByCardUUID(random_string(8)))

    def test_roles(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            # add roles
            role_repo = empty_db.getRoleRepository(session)
            role_repo.create(Role(role_id=0, role_name="admin role"))
            role_repo.create(Role(role_id=1, role_name="user role"))

            # check if roles were added
            self.assertEqual(len(role_repo.get_all()), 2)

            # rename a role
            new_name = "ÜBER ADMIN"
            r0 = role_repo.get_by_id(0)
            r0.role_name = new_name
            role_repo.update(r0)
            self.assertEqual(role_repo.get_by_id(0).role_name, new_name)

            # check linked properties
            user_repo = empty_db.getUserRepository(session)
            new_user = User(name="Mario", surname="Rossi", role_id=1)
            user_repo.create(new_user)
            role_1 = role_repo.get_by_id(1)

            # check if the role is linked to the user
            self.assertIsNotNone(role_1)
            self.assertIsNone(role_repo.get_by_id(10))
            self.assertIsNone(role_repo.get_by_role_name("nonexistent role"))

    def test_get_user_authorizations(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            userRepo = simple_db.getUserRepository(session)
            authRepo = simple_db.getAuthorizationRepository(session)
            machineRepo = simple_db.getMachineRepository(session)
            machineTypeRepo = simple_db.getMachineTypeRepository(session)
            roleRepo = simple_db.getRoleRepository(session)

            # add roles
            role = Role(role_name="normal role", authorize_all=False, maintenance=True)
            roleRepo.create(role)

            # add user with UUID
            card_UUID = random_string(8)
            user = User(name="Mario", surname="Rossi", role_id=role.role_id, card_UUID=card_UUID)
            userRepo.create(user)

            # create two machines with type 0
            machineType = MachineType(type_name="machine type: test")
            machineTypeRepo.create(machineType)

            machine1 = Machine(machine_name="test machine super", machine_type_id=machineType.type_id)
            machineRepo.create(machine1)
            machine2 = Machine(machine_name="test machine better", machine_type_id=machineType.type_id)
            machineRepo.create(machine2)

            # create one machine with type 1
            machineType1 = MachineType(type_name="machine type: another test")
            machineTypeRepo.create(machineType1)
            machine3 = Machine(machine_name="greatest machine", machine_type_id=machineType1.type_id)
            machineRepo.create(machine3)

            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T1 User should not be authorized for machine1"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine2, user), "T1 User should not be authorized for machine2"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine3, user), "T1 User should not be authorized for machine3"
            )

            auth = Authorization(user_id=user.user_id, machine_id=machine1.machine_id)
            authRepo.create(auth)
            user.authorizations = [auth]
            userRepo.update(user)
            # Reload user from db just be sure everything is saved
            user = userRepo.get_by_id(user.user_id)

            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T2 User should be authorized for machine1"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine2, user), "T2 User should not be authorized for machine2"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine3, user), "T2 User should not be authorized for machine3"
            )

            user = userRepo.getUserByCardUUID(card_UUID)
            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T3 User should be authorized for machine1"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine2, user), "T3 User should not be authorized for machine2"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine3, user), "T3 User should not be authorized for machine3"
            )

            # Check disabled is working
            user.disabled = True
            userRepo.update(user)
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine1, user),
                "T3 User should be disabled and not authorized for machine1",
            )
            user.disabled = False
            userRepo.update(user)
            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T3 User should be authorized for machine1"
            )

            # Check deleted is working
            user.deleted = True
            userRepo.update(user)
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine1, user),
                "T3 User should be deleted and not authorized for machine1",
            )
            user.deleted = False
            userRepo.update(user)
            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T3 User should be authorized for machine1"
            )

            # More auth checks
            auth2 = Authorization(user_id=user.user_id, machine_id=machine3.machine_id)
            authRepo.create(auth2)

            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T4 User should be authorized for machine1"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine2, user), "T4 User should not be authorized for machine2"
            )
            self.assertTrue(
                userRepo.IsUserAuthorizedForMachine(machine3, user), "T4 User should be authorized for machine3"
            )

            authRepo.delete(auth)
            authRepo.delete(auth2)

            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine1, user), "T4 User should not be authorized for machine1"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine2, user), "T4 User should not be authorized for machine2"
            )
            self.assertFalse(
                userRepo.IsUserAuthorizedForMachine(machine3, user), "T4 User should not be authorized for machine3"
            )

    def test_long_add_users(self):
        db = get_empty_test_db()
        with db.getSession() as session:
            USERS = 100
            name = "Mario"
            surname = "Rossi"

            role = Role(role_name="normal role", authorize_all=False, maintenance=False)
            db.getRoleRepository(session).create(role)
            userRepo = db.getUserRepository(session)

            for _ in range(USERS):
                user = User(name=name, surname=surname, role_id=role.role_id)
                userRepo.create(user)

            self.assertEqual(len(db.getUserRepository(session).get_all()), USERS)

    def test_machines(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            TYPES = 10
            MACHINES = 10
            current_id = 0

            machineRepo = empty_db.getMachineRepository(session)
            mtypeRepo = empty_db.getMachineTypeRepository(session)

            # create machines for each type
            for i in range(1, TYPES + 1):
                NAME = random_string(10)
                mtype = MachineType(
                    type_name=NAME,
                    type_timeout_min=10,
                    grace_period_min=5,
                    access_management=MachineType.MANAGEMENT_WITH_AUTHORIZATION,
                )
                mtypeRepo.create(mtype)
                self.assertEqual(len(mtypeRepo.get_all()), i, "Machine type not added correctly")
                for _ in range(1, MACHINES + 1):
                    name = random_string(6)
                    machine = Machine(machine_name=name, machine_type_id=mtype.type_id)
                    machineRepo.create(machine)

            # test that they have been added
            self.assertEqual(MACHINES * TYPES, len(machineRepo.get_all()), "Machines not added correctly")

            for current_id in range(1, MACHINES * TYPES + 1):
                machine = machineRepo.get_by_id(current_id)
                self.assertIsNotNone(machine, f"Machine {current_id} not found")
                hours = random_between(10, 1000)
                machine.machine_hours = hours
                machineRepo.update(machine)
                # reload machine
                machine = machineRepo.get_by_id(current_id)
                self.assertEqual(machine.machine_hours, hours, "Hours not updated correctly")
                self.assertEqual(machine.machine_type.type_timeout_min, 10, "Machine Type type_timeout_min is wrong")
                self.assertEqual(machine.machine_type.grace_period_min, 5, "Machine Type grace_period_min is wrong")
                self.assertEqual(
                    machine.machine_type.access_management,
                    MachineType.MANAGEMENT_WITH_AUTHORIZATION,
                    "Machine Type timeout is wrong",
                )

            for current_id in range(1, MACHINES * TYPES + 1):
                machine = machineRepo.get_by_id(current_id)
                self.assertIsNotNone(machine, f"Machine {current_id} not found")
                # test rename
                NAME = random_string(10)
                machine.machine_name = NAME
                machineRepo.update(machine)

                self.assertEqual(
                    machineRepo.get_by_id(current_id).machine_name, NAME, "Machine name not updated correctly"
                )
                # test new type

                NEW_TYPE = randint(1, TYPES)
                machine.machine_type_id = NEW_TYPE
                machineRepo.update(machine)

                self.assertEqual(
                    NEW_TYPE, machineRepo.get_by_id(current_id).machine_type_id, "Machine type not updated correctly"
                )

                machineRepo.delete(machine)
                # test delete
                self.assertEqual(MACHINES * TYPES - current_id, len(machineRepo.get_all()), "Machine has been deleted")

            # test delete
            self.assertEqual(0, len(machineRepo.get_all()), "All machines should have been deleted")

    def test_maintenances(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            maint_repo = empty_db.getMaintenanceRepository(session)

            m_type = MachineType(type_name="test type")
            empty_db.getMachineTypeRepository(session).create(m_type)

            machine = Machine(machine_name="test machine", machine_type_id=m_type.type_id)
            empty_db.getMachineRepository(session).create(machine)

            MAINTENANCES = 100

            for x in range(1, MAINTENANCES):
                hours_between = random_between(5, 10)
                description = random_string(50)
                maint = Maintenance(
                    hours_between=hours_between, description=description, machine_id=machine.machine_id
                )
                maint_repo.create(maint)

                # check getters and setters
                maint2 = maint_repo.get_model_by_id(Maintenance, maint.maintenance_id)
                self.assertIsNotNone(maint2, "Maintenance not found")

                self.assertEqual(maint2.hours_between, hours_between, "Hours between not set correctly")
                self.assertEqual(maint2.description, description, "Description not set correctly")

                maint2.hours_between = random_between(5, 10)
                maint2.description = random_string(50)
                maint_repo.update(maint2)

            # check removal
            for x in range(1, MAINTENANCES):
                maint = maint_repo.get_model_by_id(Maintenance, x)
                maint_repo.delete(maint)

            self.assertEqual(len(maint_repo.get_all()), 0, "All maintenances should have been deleted")

    def test_machine_maintenance_interaction(self):
        empty_db = get_empty_test_db()
        MACHINE_TYPE_NAME = "drill"
        MACHINES = 3

        # create simple machine types
        with empty_db.getSession() as session:
            mtr = empty_db.getMachineTypeRepository(session)
            machine_repo = empty_db.getMachineRepository(session)
            maint_repo = empty_db.getMaintenanceRepository(session)

            mt = MachineType(type_name=MACHINE_TYPE_NAME)
            mtr.create(mt)
            self.assertEqual(len(mtr.get_all()), 1, "Machine type not added correctly")

            # create simple machine
            for x in range(MACHINES):
                m = Machine(machine_name=f"TEST{x}", machine_type_id=mt.type_id)
                machine_repo.create(m)

            # create simple maintenances
            MAINTENANCES_DESC = ["change oil", "clean mirror", "empty bin"]
            for mac in machine_repo.get_all():
                for _, md in enumerate(MAINTENANCES_DESC):
                    maint = Maintenance(
                        description=md,
                        hours_between=10,
                        machine_id=mac.machine_id,
                        lcd_message="pulire",
                        instructions_url="https://www.google.com",
                    )
                    maint_repo.create(maint)

            self.assertEqual(
                len(maint_repo.get_all()), MACHINES * len(MAINTENANCES_DESC), "Maintenance not added correctly"
            )

            for m in machine_repo.get_all():
                self.assertEqual(
                    len(m.maintenances), len(MAINTENANCES_DESC), "Machine maintenances not added correctly"
                )
                for maint in m.maintenances:
                    maint_repo.delete(maint)
                m = machine_repo.get_by_id(m.machine_id)
                self.assertEqual(len(m.maintenances), 0, "Machine maintenances not removed correctly")

            self.assertEqual(0, len(maint_repo.get_all()), "All maintenances should have been deleted")

    def test_interventions(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            int_repo = simple_db.getInterventionRepository(session)
            maint_repo = simple_db.getMaintenanceRepository(session)
            machine_repo = simple_db.getMachineRepository(session)
            user_repo = simple_db.getUserRepository(session)

            user = user_repo.get_by_id(1)
            machine = machine_repo.get_by_id(1)
            maint = maint_repo.get_model_by_id(Maintenance, 1)

            for inter in int_repo.get_all():
                int_repo.delete(inter)

            # create simple intervention
            inter = Intervention(
                maintenance_id=maint.maintenance_id,
                user_id=user.user_id,
                machine_id=machine.machine_id,
                timestamp=time(),
            )
            int_repo.create(inter)

            self.assertAlmostEqual(inter.timestamp, time(), delta=1000)

            self.assertEqual(len(int_repo.get_all()), 1, "Intervention not added correctly")
            self.assertEqual(len(maint.interventions), 1, "Maintenance intervention not added correctly")
            self.assertEqual(len(machine.interventions), 1, "Machine intervention not added correctly")
            self.assertEqual(len(user.interventions), 1, "User intervention not added correctly")

            int_repo.delete(inter)

            self.assertEqual(len(int_repo.get_all()), 0, "Intervention not deleted correctly")
            self.assertEqual(len(maint.interventions), 0, "Maintenance intervention not deleted correctly")
            self.assertEqual(len(machine.interventions), 0, "Machine intervention not deleted correctly")
            self.assertEqual(len(user.interventions), 0, "User intervention not deleted correctly")

    def test_user_use_interaction(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            use_repo = simple_db.getUseRepository(session)
            user_repo = simple_db.getUserRepository(session)
            machine_repo = simple_db.getMachineRepository(session)

            userID1 = user_repo.get_by_id(1)
            machine = machine_repo.get_by_id(1)

            # create simple use
            self.assertEqual(len(userID1.uses), 0, "No previous user records should exist")
            self.assertTrue(use_repo.startUse(machine.machine_id, user=userID1, timestamp=time(), is_replay=False))
            self.assertEqual(len(userID1.uses), 1, "User start of usage is registered correctly")

            use_repo.endUse(machine_id=machine.machine_id, user=userID1, duration_s=1, is_replay=False)

            self.assertEqual(len(userID1.uses), 1, "User usage has been closed correctly")
            self.assertEqual(len(machine.uses), 1, "Machine use not added correctly")

            use_repo.delete(userID1.uses[0])

            self.assertEqual(len(userID1.uses), 0, "User use not deleted correctly")
            self.assertEqual(len(machine.uses), 0, "Machine use not deleted correctly")

            # invalid cases
            for _ in range(100):
                self.assertTrue(
                    use_repo.startUse(machine_id=machine.machine_id, user=userID1, timestamp=time(), is_replay=False)
                )

            self.assertEqual(
                1, len(session.query(Use).filter(Use.end_timestamp.is_(None)).all()), "Only one open use should exist"
            )
            self.assertEqual(
                99,
                len(session.query(Use).filter(Use.end_timestamp.is_not(None)).all()),
                "Auto-closing orphan uses should have closed 99 uses",
            )

            for _ in range(100):
                use_repo.inUse(machine_id=machine.machine_id, user=userID1, duration_s=1)

            self.assertEqual(
                99,
                len(session.query(Use).filter(Use.end_timestamp.is_not(None)).all()),
                "In use does not duplicate records",
            )

            for _ in range(100):
                use_repo.endUse(machine_id=machine.machine_id, user=userID1, duration_s=1, is_replay=False)

            self.assertEqual(
                199,
                len(session.query(Use).filter(Use.end_timestamp.is_not(None)).all()),
                "Auto-closing orphan uses should have closed 99 uses",
            )
            self.assertEqual(
                0, len(session.query(Use).filter(Use.end_timestamp.is_(None)).all()), "No open uses should exist"
            )

    def test_user_helpers(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            use_repo = simple_db.getUseRepository(session)
            user_repo = simple_db.getUserRepository(session)
            machine_repo = simple_db.getMachineRepository(session)

            userID1 = user_repo.get_by_id(1)
            machine = machine_repo.get_by_id(1)

            # create simple use
            self.assertTrue(use_repo.startUse(machine.machine_id, user=userID1, timestamp=time(), is_replay=False))
            use_repo.endUse(machine_id=machine.machine_id, user=userID1, duration_s=1, is_replay=False)

            self.assertEqual(len(userID1.uses), 1, "User use not added correctly")
            self.assertEqual(len(machine.uses), 1, "Machine use not added correctly")

            use_repo.delete(userID1.uses[0])

            self.assertEqual(len(userID1.uses), 0, "User use not deleted correctly")
            self.assertEqual(len(machine.uses), 0, "Machine use not deleted correctly")

            # invalid cases
            for _ in range(100):
                self.assertTrue(
                    use_repo.startUse(machine_id=machine.machine_id, user=userID1, timestamp=time(), is_replay=False)
                )
                self.assertTrue(use_repo.inUse(machine_id=machine.machine_id, user=userID1, duration_s=1))

            self.assertEqual(
                1, len(session.query(Use).filter(Use.end_timestamp.is_(None)).all()), "Only one open use should exist"
            )
            self.assertEqual(
                99,
                len(session.query(Use).filter(Use.end_timestamp.is_not(None)).all()),
                "Auto-closing orphan uses should have closed 99 uses",
            )

            for _ in range(100):
                use_repo.endUse(machine_id=machine.machine_id, user=userID1, duration_s=1, is_replay=False)

            self.assertEqual(
                199,
                len(session.query(Use).filter(Use.end_timestamp.is_not(None)).all()),
                "Auto-closing orphan uses should have closed 99 uses",
            )
            self.assertEqual(
                0, len(session.query(Use).filter(Use.end_timestamp.is_(None)).all()), "No open uses should exist"
            )

    def test_use_helpers(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            mac_repo = simple_db.getMachineRepository(session)
            use_repo = simple_db.getUseRepository(session)
            user_repo = simple_db.getUserRepository(session)

            self.assertEqual(0, len(mac_repo.getCurrentlyUsedMachines()), "No machines should be in use")

            self.assertTrue(
                user_repo.IsUserAuthorizedForMachine(machine=mac_repo.get_by_id(1), user=user_repo.get_by_id(1)),
                "User should be authorized",
            )
            self.assertEqual(0, user_repo.getUserTotalTime(user_id=1), "User should have no total time")
            self.assertEqual(0, mac_repo.getRelativeUseTime(machine_id=1), "Machine should have no relative time")
            self.assertEqual(0, mac_repo.getTotalUseTime(machine_id=1), "Machine should have no total time")

            self.assertFalse(
                mac_repo.getMachineMaintenanceNeeded(machine_id=1)[0], "Machine should not need maintenance"
            )

            self.assertTrue(
                use_repo.startUse(machine_id=1, user=user_repo.get_by_id(1), timestamp=time() - 240, is_replay=False)
            )

            self.assertTrue(mac_repo.isMachineCurrentlyUsed(1), "Machine should be in use")
            self.assertEqual(1, len(mac_repo.getCurrentlyUsedMachines()), "One machine should be in use")

            use_repo.endUse(machine_id=1, user=user_repo.get_by_id(1), duration_s=60, is_replay=False)
            self.assertFalse(mac_repo.isMachineCurrentlyUsed(1), "Machine should not be in use")
            self.assertEqual(0, len(mac_repo.getCurrentlyUsedMachines()), "No machine should be in use")

            self.assertLess(59.0, user_repo.getUserTotalTime(user_id=1), "User should have >59s total time")
            self.assertLess(59.0, mac_repo.getRelativeUseTime(machine_id=1), "Machine should have >59s relative time")
            self.assertLess(59.0, mac_repo.getTotalUseTime(machine_id=1), "Machine should have >59s total time")

            # Register another use
            self.assertTrue(
                use_repo.startUse(machine_id=1, user=user_repo.get_by_id(1), timestamp=time() - 120, is_replay=False)
            )
            use_repo.endUse(machine_id=1, user=user_repo.get_by_id(1), duration_s=60, is_replay=False)

            self.assertFalse(mac_repo.isMachineCurrentlyUsed(1), "Machine should not be in use")
            self.assertEqual(0, len(mac_repo.getCurrentlyUsedMachines()), "No machine should be in use")

            self.assertLess(119.0, user_repo.getUserTotalTime(user_id=1), "User should have >119s total time")
            self.assertLess(
                119.0, mac_repo.getRelativeUseTime(machine_id=1), "Machine should have >119s relative time"
            )
            self.assertLess(119.0, mac_repo.getTotalUseTime(machine_id=1), "Machine should have >119s total time")

            # register an intervention between both uses
            int_repo = simple_db.getInterventionRepository(session)
            int_repo.create(Intervention(user_id=1, machine_id=1, timestamp=time() - 180, maintenance_id=1))

            # relative use shall be only the last use duration
            self.assertLess(59.0, mac_repo.getRelativeUseTime(machine_id=1), "Machine should have >59s relative time")
            # total use shall not be changed
            self.assertLess(119.0, mac_repo.getTotalUseTime(machine_id=1), "Machine should have >119s total time")

            # Check maintenance status
            self.assertFalse(
                mac_repo.getMachineMaintenanceNeeded(machine_id=1)[0], "Machine should not need maintenance"
            )

            maint_repo = simple_db.getMaintenanceRepository(session)
            maint = maint_repo.get_model_by_id(Maintenance, 1)
            maint.hours_between = 1 / 3600  # every second
            maint_repo.update(maint)

            # Check tuple values of getMachineMaintenanceNeeded
            self.assertTrue(mac_repo.getMachineMaintenanceNeeded(machine_id=1)[0], "Machine shall need maintenance")
            self.assertEqual(
                mac_repo.getMachineMaintenanceNeeded(machine_id=1)[1],
                maint.lcd_message,
                "Correct maintenance description shall be returned",
            )

            # Perform maintenance
            inter_repo = simple_db.getInterventionRepository(session)
            inter_repo.registerInterventionsDone(machine_id=1, user_id=1)
            self.assertFalse(
                mac_repo.getMachineMaintenanceNeeded(machine_id=1)[0], "Machine shall not need maintenance anymore"
            )

            # Check user's use history
            self.assertGreater(len(user_repo.getUserUses(user_repo.get_by_id(1))), 0, "User should have use history")

    def test_boards(self):
        simple_db = get_simple_db()
        with simple_db.getSession() as session:
            board_repo = simple_db.getBoardsRepository(session)
            machine_repo = simple_db.getMachineRepository(session)
            for mac in machine_repo.get_all():
                board_repo.registerBoard(f"1.2.3.{mac.machine_id}", "0.1.2", f"SN{mac.machine_id}", 300000, mac)

            # check if boards were added
            self.assertEqual(len(board_repo.get_all()), len(machine_repo.get_all()))

            for mac in machine_repo.get_all():
                board_repo.registerBoard(f"1.3.4.{mac.machine_id}", "0.1.3", f"SN{mac.machine_id}", 300000, mac)

            # check no duplicates were created
            self.assertEqual(len(board_repo.get_all()), len(machine_repo.get_all()))

            # check last_seen is updated
            for board in board_repo.get_all():
                self.assertAlmostEqual(board.last_seen, time(), delta=1000)


if __name__ == "__main__":
    configure_logger()
    unittest.main()



""" Test the database backend. """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import unittest
import os

# Import Module
import shutil

from FabOMatic.database.DatabaseBackend import DatabaseBackend, getDatabaseUrl
from tests.common import TEST_SETTINGS_PATH, get_empty_test_db, configure_logger


class TestMigrations(unittest.TestCase):
    def test_migrations(self):
        src_dir = os.path.join(os.path.dirname(__file__), "databases")
        for file in os.listdir(src_dir):
            source = os.path.abspath(os.path.join(src_dir, file))
            dest = getDatabaseUrl(TEST_SETTINGS_PATH)
            if dest.startswith("sqlite:///"):
                # Remove the prefix to get the file path
                dest = dest[len("sqlite:///") :]

            self.assertTrue(os.path.exists(source))
            # Overwrite any existing database with the previous version
            shutil.copyfile(source, dest)

            # Now check that the upgrade works
            self.performUpgrade(file)

    def performUpgrade(self, file: str):
        backend = DatabaseBackend(TEST_SETTINGS_PATH)
        print("Testing upgrade for file : ", file)
        try:
            backend.createAndUpdateDatabase()
            with backend.getSession() as session:
                user_repo = backend.getUserRepository(session)
                self.assertGreaterEqual(len(user_repo.get_all()), 1)
                role_repo = backend.getRoleRepository(session)
                self.assertGreaterEqual(len(role_repo.get_all()), 3)
                machine_repo = backend.getMachineRepository(session)
                self.assertGreaterEqual(len(machine_repo.get_all()), 1)
                type_repo = backend.getMachineTypeRepository(session)
                self.assertGreaterEqual(len(type_repo.get_all()), 1)
                maint_repo = backend.getMaintenanceRepository(session)
                self.assertGreaterEqual(len(maint_repo.get_all()), 0)
                inter_repo = backend.getInterventionRepository(session)
                self.assertGreaterEqual(len(inter_repo.get_all()), 0)

        except Exception as e:
            self.fail(f"Failed to upgrade the database {file} : {e}")



""" Test the database backend. """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import unittest

from FabOMatic.database.DatabaseBackend import DatabaseBackend
from FabOMatic.database.models import User, Role
from tests.common import TEST_SETTINGS_PATH, get_empty_test_db, configure_logger


class TestAuth(unittest.TestCase):
    def test_connection(self):
        _ = DatabaseBackend(TEST_SETTINGS_PATH)

    def test_user_auth(self):
        empty_db = get_empty_test_db()
        with empty_db.getSession() as session:
            empty_db.getRoleRepository(session).create(Role(role_id=1, role_name="admin"))

            test_users = [
                ("Alessandro", "Rossi", "pass1"),
                ("Lorenzo", "Bianchi", "pass2"),
                ("Diego", "Verdi", "pass3"),
                ("Tommaso", "Colombo", "pass4"),
                ("Riccardo", "Fumagalli", "pass5"),
            ]

            user_repo = empty_db.getUserRepository(session)
            # add roles
            for n, s, p in test_users:
                usr = User(name=n, surname=s, role_id=1)
                usr.set_password(p)
                usr.email = f"{n}.{s}@fablabbg.org"

                user_repo.create(usr)
                reloaded = user_repo.get_by_id(usr.user_id)
                self.assertEqual(reloaded.name, n)
                self.assertEqual(reloaded.surname, s)
                self.assertEqual(reloaded.role_id, 1)
                self.assertTrue(reloaded.check_password(p), "Password not set correctly")
                self.assertFalse(reloaded.check_password("wrong_password"), "Password not verified correctly")
                self.assertEqual(reloaded.email, usr.email)
                self.assertGreater(len(reloaded.serialize()), 0)

                # Token verification

                key = b"192b9b237822dd23"
                wrong_key = b"192b9b237822dd24"
                salt = b"fablab-bg"

                token = reloaded.get_reset_token(key, salt)
                self.assertEqual(
                    reloaded.user_id,
                    User.verify_reset_token(token, key, salt, 1),
                    f"Token not verified correctly for user {reloaded}",
                )
                self.assertIsNone(
                    User.verify_reset_token(token, wrong_key, salt, 1),
                    f"Token not verified correctly for user {reloaded}",
                )
                self.assertIsNone(
                    User.verify_reset_token(token, wrong_key, salt, -1),
                    f"Token expiration not verified correctly for user {reloaded}",
                )

                token = "PROVA"
                self.assertIsNone(User.verify_reset_token(token, key, salt), "Token not verified correctly")


if __name__ == "__main__":
    configure_logger()
    unittest.main()



""" This module contains the routes for the maintenance management. """

# pylint: disable=C0116

import logging
import os

from flask import render_template, request, redirect, url_for, flash
from flask_login import login_required
from flask_babel import gettext
from werkzeug.utils import secure_filename

from FabOMatic.database.models import Machine, Maintenance
from .webapplication import DBSession, allowed_file, app, UPLOAD_FOLDER


@app.route("/maintenances")
@login_required
def maintenances():
    machine_filter = request.args.get("machine")
    description_filter = request.args.get("description")

    session = DBSession()
    query = session.query(Maintenance).join(Maintenance.machine)

    if machine_filter and machine_filter.isdigit():
        query = query.filter(Machine.machine_id == int(machine_filter))
    if description_filter:
        query = query.filter(Maintenance.description.contains(description_filter))

    maintenances = query.all()
    machines = session.query(Machine).all()

    return render_template("view_maintenances.html", maintenances=maintenances, machines=machines)


@app.route("/maintenances/add", methods=["GET", "POST"])
@login_required
def add_maintenance():
    session = DBSession()
    logging.debug("Processing add_maintenance %s", request)
    if request.method == "POST":
        hours_between = request.form["hours_between"]
        try:
            if float(hours_between) <= 0:
                flash(gettext("Hours between must be a positive number."))
                return redirect(url_for("add_maintenance"))
            hours_between = float(hours_between)
        except ValueError:
            flash(gettext("Hours between must be a number."))
            return redirect(url_for("add_maintenance"))

        description = request.form["description"]
        machine_id = request.form["machine_id"]
        lcd_message = request.form["lcd_message"]
        url = request.form["instructions_url"]

        maintenance = Maintenance(
            hours_between=hours_between,
            description=description,
            machine_id=machine_id,
            lcd_message=lcd_message,
            instructions_url=url,
        )
        session.add(maintenance)
        session.commit()
        return redirect(url_for("maintenances"))

    machines = session.query(Machine).all()
    return render_template("add_maintenance.html", machines=machines)


@app.route("/maintenances/edit/<int:maintenance_id>", methods=["GET", "POST"])
@login_required
def edit_maintenance(maintenance_id):
    session = DBSession()
    maintenance = session.query(Maintenance).filter_by(maintenance_id=maintenance_id).one()
    if request.method == "POST":
        hours_between = request.form["hours_between"]
        maintenance.description = request.form["description"]
        maintenance.machine_id = request.form["machine_id"]
        maintenance.lcd_message = request.form["lcd_message"]
        maintenance.instructions_url = request.form["instructions_url"]
        try:
            if float(hours_between) <= 0:
                flash(gettext("Hours between must be a positive number."))
                return redirect(url_for("edit_maintenance", maintenance_id=maintenance_id))
            hours_between = float(hours_between)
        except ValueError:
            flash(gettext("Hours between must be a number."))
            return redirect(url_for("edit_maintenance", maintenance_id=maintenance_id))

        maintenance.hours_between = hours_between

        session.add(maintenance)
        session.commit()
        return redirect(url_for("maintenances"))
    else:
        machines = session.query(Machine).all()
        return render_template(
            "edit_maintenance.html",
            machines=machines,
            maintenance_id=maintenance.maintenance_id,
            maintenance=maintenance,
        )


@app.route("/maintenances/delete/<int:maintenance_id>", methods=["GET", "POST"])
def delete_maintenance(maintenance_id):
    session = DBSession()
    maintenance = session.query(Maintenance).filter_by(maintenance_id=maintenance_id).one()
    if not maintenance:
        return gettext("Maintenance not found"), 404

    if request.method == "POST":
        session.delete(maintenance)
        session.commit()
        flash(gettext("Maintenance deleted successfully."))
        return redirect(url_for("maintenances"))

    return render_template("delete_maintenance.html", maintenance=maintenance)



""" This module contains the routes for the uses pages. """

# pylint: disable=C0116

from datetime import datetime

from flask import render_template, request, redirect, url_for, flash
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import Machine, Use, User
from FabOMatic.database.repositories import MachineRepository, UserRepository
from .webapplication import DBSession, app, excel


@app.route("/machines/history/<int:machine_id>", methods=["GET"])
@login_required
def view_machine_use_history(machine_id):
    session = DBSession()
    uses = session.query(Use).filter_by(machine_id=machine_id).order_by(Use.start_timestamp.desc()).limit(500).all()
    machine = session.query(Machine).filter_by(machine_id=machine_id).one()
    if machine is None:
        return gettext("Machine not found"), 404

    return render_template("view_machine_use_history.html", uses=uses, machine=machine)


@app.route("/delete_use/<int:use_id>", methods=["POST", "GET"])
@login_required
def delete_use(use_id):
    session = DBSession()
    use = session.query(Use).filter_by(use_id=use_id).one()
    # ASk for confirmation before deleting
    if request.method == "GET":
        return render_template("delete_use.html", use=use)

    if use:
        # Correct machine cumulated hours if requested.
        correct = request.form.get("correctTotal", type=str, default="")
        if correct == "on":
            duration = use.end_timestamp - use.start_timestamp
            machine = session.query(Machine).filter_by(machine_id=use.machine_id).one()
            if machine and duration > 0:
                machine.machine_hours -= duration / 3600.0

        session.delete(use)
        session.commit()
        flash(gettext("Use deleted successfully."))
    else:
        return "Use not found.", 404
    return redirect(url_for("view_uses"))


@app.route("/view_uses", methods=["GET"])
@login_required
def view_uses():
    session = DBSession()
    user_id = request.args.get("user_id")
    machine_id = request.args.get("machine_id")
    start_time = request.args.get("start_time")

    # Convert start_time to a datetime object
    if start_time:
        start_time = datetime.strptime(start_time, "%Y-%m-%dT%H:%M")

    # Query the database to get the data
    uses = session.query(Use)

    if user_id and user_id.isdigit():
        user_id = int(user_id)
        uses = uses.filter(Use.user_id == user_id)

    if machine_id and machine_id.isdigit():
        machine_id = int(machine_id)
        uses = uses.filter(Use.machine_id == machine_id)

    if start_time:
        uses = uses.filter(Use.start_timestamp >= start_time)

    uses = uses.order_by(Use.start_timestamp.desc()).limit(500).all()

    # Query the database to get all users and machines for the filter dropdowns
    all_users = session.query(User).filter_by(deleted=False).all()
    all_machines = session.query(Machine).all()

    return render_template(
        "view_uses.html",
        uses=uses,
        all_users=all_users,
        all_machines=all_machines,
        selected_user_id=user_id,
        selected_machine_id=machine_id,
        selected_start_time=start_time,
    )


@app.route("/add_use", methods=["GET"])
@login_required
def add_use():
    with DBSession() as session:
        users = session.query(User).filter_by(deleted=False).order_by(User.name).all()
        machines = session.query(Machine).order_by(Machine.machine_name).all()
        return render_template("add_use.html", users=users, machines=machines)


@app.route("/add_use", methods=["POST"])
@login_required
def add_use_post():
    with DBSession() as session:
        use_data = request.form
        user_id = use_data["user_id"]
        machine_id = use_data["machine_id"]
        try:
            start_timestamp = datetime.strptime(use_data["start_timestamp"], "%Y-%m-%dT%H:%M")
            start_timestamp = start_timestamp.timestamp()
        except ValueError:
            flash(gettext("Invalid start timestamp."), "error")
            return redirect(url_for("add_use"))

        try:
            end_timestamp = datetime.strptime(use_data["end_timestamp"], "%Y-%m-%dT%H:%M")
            end_timestamp = end_timestamp.timestamp()
        except ValueError:
            end_timestamp = None

        if end_timestamp and end_timestamp < start_timestamp:
            flash(gettext("End timestamp cannot be before start timestamp."), "error")
            return redirect(url_for("add_use"))

        machine_repo = MachineRepository(session)
        machine = machine_repo.get_by_id(machine_id)
        if machine is None:
            flash(gettext("Machine not found."), "error")
            return redirect(url_for("add_use"))

        user_repo = UserRepository(session)
        user = user_repo.get_by_id(user_id)
        if user is None:
            flash(gettext("User not found."), "error")
            return redirect(url_for("add_use"))

        new_use = Use(
            user_id=user_id,
            machine_id=machine_id,
            start_timestamp=start_timestamp,
            last_seen=start_timestamp,
            end_timestamp=end_timestamp,
        )

        # Update machine usage
        machine.machine_hours += (end_timestamp - start_timestamp) / 3600.0

        session.add(new_use)
        session.commit()
        flash(gettext("Registration added successfully."))
        return redirect(url_for("view_uses"))


@app.route("/uses/export", methods=["GET"])
@login_required
def uses_export():
    session = DBSession()
    return excel.make_response_from_tables(session, [Use, Machine, User], "xlsx", file_name="uses")


@app.route("/machines/history/<int:machine_id>/export", methods=["GET"])
@login_required
def machine_use_export(machine_id):
    session = DBSession()
    machine = session.query(Machine).filter_by(machine_id=machine_id).one()
    if machine is None:
        return gettext("Machine not found"), 404
    uses = session.query(Use).filter_by(machine_id=machine_id).order_by(Use.start_timestamp.desc()).all()
    columns = ["use_id", "user_id", "machine_id", "start_timestamp", "end_timestamp"]
    return excel.make_response_from_query_sets(uses, columns, "xlsx", file_name="machine_use")



""" This is the main file of the project. It is used to run the project. """

import getopt
import sys
import FabOMatic
import logging


def getArgLoglevel():
    try:
        options, args = getopt.getopt(sys.argv[1:], "l:", ["loglevel ="])
    except:
        print("Error Message ")

    for name, value in options:
        if name in ["-l", "--loglevel"]:
            return int(value)
    return logging.INFO


try:
    FabOMatic.start(getArgLoglevel())
except (KeyboardInterrupt, SystemExit):
    logging.info("Exiting...")
    sys.exit()



""" This module contains the routes for the users management. """

# pylint: disable=C0116

import re

from flask import render_template, request, redirect, url_for, flash
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import User, Role, UnknownCard
from FabOMatic.web.authentication import send_reset_email
from .webapplication import DBSession, app, excel


@app.route("/users", methods=["GET"])
@login_required
def view_users():
    session = DBSession()
    users = session.query(User).filter_by(deleted=False).order_by(User.user_id).all()
    cards = session.query(UnknownCard).order_by(UnknownCard.timestamp.desc()).limit(20)
    return render_template("view_users.html", users=users, cards=cards)


@app.route("/users/add", methods=["GET"])
@login_required
def add_user():
    session = DBSession()
    roles = session.query(Role).order_by(Role.role_id).all()
    uuid = request.args.get("card_uuid", None)
    return render_template("add_user.html", roles=roles, card_UUID=uuid)


@app.route("/users/reset/<int:user_id>", methods=["GET", "POST"])
@login_required
def reset_user(user_id):
    session = DBSession()
    user = session.query(User).filter_by(user_id=user_id).one()
    if not user:
        return gettext("User not found"), 404

    if request.method == "POST":
        send_reset_email(user)
        flash(gettext("An email has been sent with instructions to reset password."), "info")
        return redirect(url_for("view_users"))

    return render_template("reset_user.html", user=user)


@app.route("/users/create", methods=["POST"])
@login_required
def create_user():
    session = DBSession()
    user_data = request.form
    card_UUID = user_data.get("card_UUID", None)

    if card_UUID == "" or card_UUID == "None":  # If the card_UUID is empty, set it to None
        card_UUID = None

    if card_UUID and not re.match(r"^[0-9A-Fa-f]{8}$", card_UUID):
        flash(gettext("Invalid card UUID. Please enter either 8 hexadecimal characters or leave it empty."), "error")
        return redirect(url_for("view_users"))

    check_user = session.query(User).filter(User.card_UUID.isnot(None)).filter_by(card_UUID=card_UUID).one_or_none()
    if check_user:
        flash(
            f"This card ID ({card_UUID}) is already assigned to another user ({check_user.name} {check_user.surname})",
            "error",
        )
        return redirect(url_for("view_users"))
    new_user = User(
        name=user_data["name"],
        surname=user_data["surname"],
        role_id=user_data["role_id"],
        disabled=user_data.get("disabled", "off") == "on",
        card_UUID=card_UUID,
        email=user_data["email"],
    )
    session.add(new_user)

    # Delete the card from the unknown cards table if it exists
    for card in session.query(UnknownCard).filter_by(card_UUID=card_UUID).all():
        session.delete(card)

    session.commit()

    if new_user.role.backend_admin and len(new_user.email) == 0:
        flash(
            gettext(
                "You have created a backend admin user without an email address. User will not be able to log on."
            ),
            "warning",
        )
    return redirect(url_for("view_users"))


@app.route("/users/edit/<int:user_id>", methods=["GET"])
@login_required
def edit_user(user_id):
    session = DBSession()
    user = session.query(User).filter_by(user_id=user_id).one()
    roles = session.query(Role).order_by(Role.role_id).all()
    if user:
        return render_template("edit_user.html", user=user, roles=roles)
    else:
        return gettext("User not found"), 404


@app.route("/users/update", methods=["POST"])
@login_required
def update_user():
    session = DBSession()
    user_data = request.form
    user = session.query(User).filter_by(user_id=user_data["user_id"]).one()
    if user:
        # Validate the card_UUID
        card_UUID = user_data.get("card_UUID", None)

        if card_UUID and not re.match(r"^[0-9A-Fa-f]{8}$", card_UUID):
            flash(
                gettext("Invalid card UUID. Please enter either 8 hexadecimal characters or leave it empty."), "error"
            )
            return redirect(url_for("edit_user", user_id=user.user_id))
        if card_UUID == "" or card_UUID == "None":
            card_UUID = None

        check_user = (
            session.query(User).filter(User.card_UUID.isnot(None)).filter_by(card_UUID=card_UUID).one_or_none()
        )
        if check_user and check_user.user_id != user.user_id:
            flash(
                gettext("This card ID is already assigned to another user")
                + f" ({check_user.name} {check_user.surname})",
                "error",
            )
            return redirect(url_for("edit_user", user_id=user.user_id))
        user.name = user_data["name"]
        user.surname = user_data["surname"]
        user.role_id = user_data["role_id"]
        user.card_UUID = card_UUID
        user.disabled = user_data.get("disabled", "off") == "on"
        user.email = user_data["email"]
        session.commit()
        return redirect(url_for("view_users"))
    else:
        return gettext("User not found"), 404


@app.route("/users/delete/<int:user_id>", methods=["GET", "POST"])
@login_required
def delete_user(user_id):
    session = DBSession()
    user = session.query(User).filter_by(user_id=user_id).one()
    if not user:
        return gettext("User not found"), 404

    if request.method == "POST":
        user.deleted = True
        user.card_UUID = None
        session.commit()
        return redirect(url_for("view_users"))

    return render_template("delete_user.html", user=user)


@app.route("/users/export", methods=["GET"])
@login_required
def users_export():
    session = DBSession()
    return excel.make_response_from_tables(session, [User, Role], "xlsx", file_name="users")
